from asynctest import TestCase

from parameterized import parameterized

from async_hvac import AsyncClient
from async_hvac.tests.util import RequestsMocker


class TestGcpMethods(TestCase):
    """Unit tests providing coverage for GCP auth backend-related methods/routes."""

    @parameterized.expand([
        ("default mount point", "custom_role", "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9", None),
        ("custom mount point", "custom_role", "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9", "gcp-not-default")
    ])
    @RequestsMocker()
    async def test_auth_gcp(self, test_label, test_role, test_jwt, mount_point, requests_mocker):
        mock_response = {
            'auth': {
                'accessor': 'accessor-1234-5678-9012-345678901234',
                'client_token': 'cltoken-1234-5678-9012-345678901234',
                'lease_duration': 10000,
                'metadata': {
                    'role': 'custom_role',
                    'service_account_email': 'dev1@project-123456.iam.gserviceaccount.com',
                    'service_account_id': '111111111111111111111'
                },
                'policies': [
                    'default',
                    'custom_role'
                ],
                'renewable': True
            },
            'data': None,
            'lease_duration': 0,
            'lease_id': '',
            'renewable': False,
            'request_id': 'requesti-1234-5678-9012-345678901234',
            'warnings': [],
            'wrap_info': None
        }
        mock_url = 'http://127.0.0.1:8200/v1/auth/{0}/login'.format(
                'gcp' if mount_point is None else mount_point)
        requests_mocker.register_uri(
            method='POST',
            url=mock_url,
            json=mock_response
        )
        client = AsyncClient()

        if mount_point is None:
            actual_response = await client.auth_gcp(
                role=test_role,
                jwt=test_jwt
            )
        else:
            actual_response = await client.auth_gcp(
                role=test_role,
                jwt=test_jwt,
                mount_point=mount_point
            )

        # ensure we received our mock response data back successfully
        self.assertEqual(mock_response, actual_response)
        await client.close()



from asynctest import TestCase

from async_hvac import AsyncClient
from async_hvac.tests.util import RequestsMocker


class TestAuthMethods(TestCase):
    """Tests for methods related to Vault sys/auth routes."""

    @RequestsMocker()
    async def test_tune_auth_backend(self, requests_mocker):
        expected_status_code = 204
        test_backend_type = 'approle'
        test_mount_point = 'approle-test'
        test_description = 'this is a test description'
        requests_mocker.register_uri(
            method='POST',
            url='http://127.0.0.1:8200/v1/sys/auth/{0}/tune'.format(test_mount_point),
            status_code=expected_status_code,
        )
        client = AsyncClient()
        actual_response = await client.tune_auth_backend(
            backend_type=test_backend_type,
            mount_point=test_mount_point,
            description=test_description,
        )

        self.assertEqual(
            first=expected_status_code,
            second=actual_response.status,
        )

        actual_request_params = requests_mocker.requests[
            ('post', 'http://127.0.0.1:8200/v1/sys/auth/{0}/tune'.format(test_mount_point))][0].kwargs['json']

        # Ensure we sent through an optional tune parameter as expected
        self.assertEqual(
            first=test_description,
            second=actual_request_params['description'],
        )
        await client.close()

    @RequestsMocker()
    async def test_get_auth_backend_tuning(self, requests_mocker):
        expected_status_code = 200
        test_backend_type = 'approle'
        test_mount_point = 'approle-test'
        mock_response = {
            'max_lease_ttl': 12345678,
            'lease_id': '',
            'force_no_cache': False,
            'warnings': None,
            'data': {
                'force_no_cache': False,
                'default_lease_ttl': 2764800,
                'max_lease_ttl': 12345678
            },
            'wrap_info': None,
            'auth': None,
            'lease_duration': 0,
            'request_id': '673f2336-3235-b988-2194-c68261a02bfe',
            'default_lease_ttl': 2764800,
            'renewable': False
        }
        requests_mocker.register_uri(
            method='GET',
            url='http://127.0.0.1:8200/v1/sys/auth/{0}/tune'.format(test_mount_point),
            status_code=expected_status_code,
            json=mock_response
        )
        client = AsyncClient()
        actual_response = await client.get_auth_backend_tuning(
            backend_type=test_backend_type,
            mount_point=test_mount_point,
        )

        self.assertEqual(
            first=mock_response,
            second=actual_response,
        )
        await client.close()



import asyncio
import asynctest
import binascii
import sys
from base64 import b64decode
from uuid import UUID

from async_hvac import AsyncClient, Client, exceptions
from async_hvac.tests import util

loop = asyncio.get_event_loop()


def create_client(sync=False, **kwargs):
    if sync:
        return Client(url='https://127.0.0.1:8200',
                      cert=('test/client-cert.pem', 'test/client-key.pem'),
                      verify='test/server-cert.pem',
                      loop=IntegrationTest.loop,
                      **kwargs)
    else:
        return AsyncClient(url='https://127.0.0.1:8200',
                           cert=('test/client-cert.pem', 'test/client-key.pem'),
                           verify='test/server-cert.pem',
                           loop=IntegrationTest.loop,
                           **kwargs)


class IntegrationTest(asynctest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.manager = util.ServerManager(
            config_path='test/vault-tls.hcl',
            client=create_client(sync=True))
        cls.manager.start()
        cls.manager.initialize()
        cls.manager.unseal()

    @classmethod
    def tearDownClass(cls):
        cls.manager.stop()

    def root_token(self):
        cls = type(self)
        return cls.manager.root_token

    async def setUp(self):
        self.client = create_client(token=self.root_token())

    async def tearDown(self):
        await self.client.close()

    async def test_verifiy_false(self):
        await self.client.close()
        self.client = AsyncClient(
            url='https://127.0.0.1:8200', verify=False, loop=IntegrationTest.loop)
        assert 'ha_enabled' in (await self.client.ha_status)

    async def test_unseal_multi(self):
        cls = type(self)

        await self.client.seal()

        keys = cls.manager.keys

        result = await self.client.unseal_multi(keys[0:2])

        assert result['sealed']
        assert result['progress'] == 2

        result = await self.client.unseal_reset()
        assert result['progress'] == 0
        result = await self.client.unseal_multi(keys[1:3])
        assert result['sealed']
        assert result['progress'] == 2
        result = await self.client.unseal_multi(keys[0:1])
        result = await self.client.unseal_multi(keys[2:3])
        assert not result['sealed']

    async def test_seal_unseal(self):
        cls = type(self)

        assert not (await self.client.is_sealed())

        await self.client.seal()

        assert (await self.client.is_sealed())

        cls.manager.unseal()

        assert not (await self.client.is_sealed())

    async def test_ha_status(self):
        assert 'ha_enabled' in (await self.client.ha_status)

    async def test_generic_secret_backend(self):
        await self.client.write('secret/foo', zap='zip')
        result = await self.client.read('secret/foo')

        assert result['data']['zap'] == 'zip'

        await self.client.delete('secret/foo')

    async def test_list_directory(self):
        await self.client.write('secret/test-list/bar/foo', value='bar')
        await self.client.write('secret/test-list/foo', value='bar')
        result = await self.client.list('secret/test-list')

        assert result['data']['keys'] == ['bar/', 'foo']

        await self.client.delete('secret/test-list/bar/foo')
        await self.client.delete('secret/test-list/foo')

    async def test_write_with_response(self):
        if 'transit/' in (await self.client.list_secret_backends()):
            await self.client.disable_secret_backend('transit')
        await self.client.enable_secret_backend('transit')

        plaintext = 'test'

        await self.client.write('transit/keys/foo')

        result = await self.client.write('transit/encrypt/foo', plaintext=plaintext)
        ciphertext = result['data']['ciphertext']

        result = await self.client.write('transit/decrypt/foo', ciphertext=ciphertext)
        assert result['data']['plaintext'] == plaintext

    async def test_wrap_write(self):
        if 'approle/' not in (await self.client.list_auth_backends()):
            await self.client.enable_auth_backend("approle")

        await self.client.write("auth/approle/role/testrole")
        result = await self.client.write('auth/approle/role/testrole/secret-id', wrap_ttl="10s")
        assert 'token' in result['wrap_info']
        await self.client.unwrap(result['wrap_info']['token'])
        await self.client.disable_auth_backend("approle")

    async def test_read_nonexistent_key(self):
        assert not (await self.client.read('secret/I/dont/exist'))

    async def test_auth_backend_manipulation(self):
        assert 'github/' not in (await self.client.list_auth_backends())

        await self.client.enable_auth_backend('github')
        assert 'github/' in (await self.client.list_auth_backends())

        self.client.token = self.root_token()
        await self.client.disable_auth_backend('github')
        assert 'github/' not in (await self.client.list_auth_backends())

    async def test_secret_backend_manipulation(self):
        assert 'test/' not in (await self.client.list_secret_backends())

        await self.client.enable_secret_backend('generic', mount_point='test')
        assert 'test/' in (await self.client.list_secret_backends())

        secret_backend_tuning = await self.client.get_secret_backend_tuning('generic', mount_point='test')
        self.assertEqual(secret_backend_tuning['max_lease_ttl'], 2764800)
        self.assertEqual(secret_backend_tuning['default_lease_ttl'], 2764800)

        await self.client.tune_secret_backend('generic', mount_point='test', default_lease_ttl='3600s', max_lease_ttl='8600s')
        secret_backend_tuning = await self.client.get_secret_backend_tuning('generic', mount_point='test')

        assert 'max_lease_ttl' in secret_backend_tuning
        self.assertEqual(secret_backend_tuning['max_lease_ttl'], 8600)
        assert 'default_lease_ttl' in secret_backend_tuning
        self.assertEqual(secret_backend_tuning['default_lease_ttl'], 3600)

        await self.client.remount_secret_backend('test', 'foobar')
        assert 'test/' not in (await self.client.list_secret_backends())
        assert 'foobar/' in (await self.client.list_secret_backends())

        self.client.token = self.root_token()
        await self.client.disable_secret_backend('foobar')
        assert 'foobar/' not in (await self.client.list_secret_backends())

    async def test_audit_backend_manipulation(self):
        assert 'tmpfile/' not in (await self.client.list_audit_backends())

        options = {
            'path': '/tmp/vault.audit.log'
        }

        await self.client.enable_audit_backend('file', options=options, name='tmpfile')
        assert 'tmpfile/' in (await self.client.list_audit_backends())

        self.client.token = self.root_token()
        await self.client.disable_audit_backend('tmpfile')
        assert 'tmpfile/' not in (await self.client.list_audit_backends())

    async def prep_policy(self, name):
        text = """
        path "sys" {
            policy = "deny"
        }
        path "secret" {
            policy = "write"
        }
        """
        obj = {
            'path': {
                'sys': {
                    'policy': 'deny'},
                'secret': {
                    'policy': 'write'}
            }
        }
        await self.client.set_policy(name, text)
        return text, obj

    async def test_policy_manipulation(self):
        assert 'root' in (await self.client.list_policies())
        assert (await self.client.get_policy('test')) is None
        policy, parsed_policy = await self.prep_policy('test')
        assert 'test' in (await self.client.list_policies())
        assert policy == (await self.client.get_policy('test'))
        assert parsed_policy == (await self.client.get_policy('test', parse=True))

        await self.client.delete_policy('test')
        assert 'test' not in (await self.client.list_policies())

    async def test_json_policy_manipulation(self):
        assert 'root' in (await self.client.list_policies())

        policy = {
            "path": {
                "sys": {
                    "policy": "deny"
                },
                "secret": {
                    "policy": "write"
                }
            }
        }

        await self.client.set_policy('test', policy)
        assert 'test' in (await self.client.list_policies())

        await self.client.delete_policy('test')
        assert 'test' not in (await self.client.list_policies())

    async def test_auth_token_manipulation(self):
        result = await self.client.create_token(lease='1h', renewable=True)
        assert result['auth']['client_token']

        lookup = await self.client.lookup_token(result['auth']['client_token'])
        assert result['auth']['client_token'] == lookup['data']['id']

        renew = await self.client.renew_token(lookup['data']['id'])
        assert result['auth']['client_token'] == renew['auth']['client_token']

        await self.client.revoke_token(lookup['data']['id'])

        try:
            lookup = await self.client.lookup_token(result['auth']['client_token'])
            assert False
        except exceptions.Forbidden:
            assert True
        except exceptions.InvalidPath:
            assert True
        except exceptions.InvalidRequest:
            assert True

    async def test_userpass_auth(self):
        if 'userpass/' in (await self.client.list_auth_backends()):
            await self.client.disable_auth_backend('userpass')

        await self.client.enable_auth_backend('userpass')

        await self.client.write('auth/userpass/users/testuser', password='testpass', policies='not_root')

        result = await self.client.auth_userpass('testuser', 'testpass')

        assert self.client.token == result['auth']['client_token']
        assert (await self.client.is_authenticated())

        self.client.token = self.root_token()
        await self.client.disable_auth_backend('userpass')

    async def test_create_userpass(self):
        if 'userpass/' not in (await self.client.list_auth_backends()):
            await self.client.enable_auth_backend('userpass')

        await self.client.create_userpass('testcreateuser', 'testcreateuserpass', policies='not_root')

        result = await self.client.auth_userpass('testcreateuser', 'testcreateuserpass')

        assert self.client.token == result['auth']['client_token']
        assert (await self.client.is_authenticated())

        # Test ttl:
        self.client.token = self.root_token()
        await self.client.create_userpass('testcreateuser', 'testcreateuserpass', policies='not_root', ttl='10s')
        self.client.token = result['auth']['client_token']

        result = await self.client.auth_userpass('testcreateuser', 'testcreateuserpass')

        assert result['auth']['lease_duration'] == 10

        self.client.token = self.root_token()
        await self.client.disable_auth_backend('userpass')

    async def test_list_userpass(self):
        if 'userpass/' not in (await self.client.list_auth_backends()):
            await self.client.enable_auth_backend('userpass')

        # add some users and confirm that they show up in the list
        await self.client.create_userpass('testuserone', 'testuseronepass', policies='not_root')
        await self.client.create_userpass('testusertwo', 'testusertwopass', policies='not_root')

        user_list = await self.client.list_userpass()
        assert 'testuserone' in user_list['data']['keys']
        assert 'testusertwo' in user_list['data']['keys']

        # delete all the users and confirm that list_userpass() doesn't fail
        for user in user_list['data']['keys']:
            await self.client.delete_userpass(user)

        no_users_list = await self.client.list_userpass()
        assert no_users_list is None

    async def test_read_userpass(self):
        if 'userpass/' not in (await self.client.list_auth_backends()):
            await self.client.enable_auth_backend('userpass')

        # create user to read
        await self.client.create_userpass('readme', 'mypassword', policies='not_root')

        # test that user can be read
        read_user = await self.client.read_userpass('readme')
        assert 'not_root' in read_user['data']['policies']

        # teardown
        await self.client.disable_auth_backend('userpass')

    async def test_update_userpass_policies(self):
        if 'userpass/' not in (await self.client.list_auth_backends()):
            await self.client.enable_auth_backend('userpass')

        # create user and then update its policies
        await self.client.create_userpass('updatemypolicies', 'mypassword', policies='not_root')
        await self.client.update_userpass_policies('updatemypolicies', policies='somethingelse')

        # test that policies have changed
        updated_user = await self.client.read_userpass('updatemypolicies')
        assert 'somethingelse' in updated_user['data']['policies']

        # teardown
        await self.client.disable_auth_backend('userpass')

    async def test_update_userpass_password(self):
        if 'userpass/' not in (await self.client.list_auth_backends()):
            await self.client.enable_auth_backend('userpass')

        # create user and then change its password
        await self.client.create_userpass('changeme', 'mypassword', policies='not_root')
        await self.client.update_userpass_password('changeme', 'mynewpassword')

        # test that new password authenticates user
        result = await self.client.auth_userpass('changeme', 'mynewpassword')
        assert self.client.token == result['auth']['client_token']
        assert (await self.client.is_authenticated())

        # teardown
        self.client.token = self.root_token()
        await self.client.disable_auth_backend('userpass')

    async def test_delete_userpass(self):
        if 'userpass/' not in (await self.client.list_auth_backends()):
            await self.client.enable_auth_backend('userpass')

        await self.client.create_userpass('testcreateuser', 'testcreateuserpass', policies='not_root')

        result = await self.client.auth_userpass('testcreateuser', 'testcreateuserpass')

        assert self.client.token == result['auth']['client_token']
        assert (await self.client.is_authenticated())

        self.client.token = self.root_token()
        await self.client.delete_userpass('testcreateuser')
        with self.assertRaises(exceptions.InvalidRequest):
            await self.client.auth_userpass('testcreateuser', 'testcreateuserpass')

    async def test_app_id_auth(self):
        if 'app-id/' in (await self.client.list_auth_backends()):
            await self.client.disable_auth_backend('app-id')

        await self.client.enable_auth_backend('app-id')

        await self.client.write('auth/app-id/map/app-id/foo', value='not_root')
        await self.client.write('auth/app-id/map/user-id/bar', value='foo')

        result = await self.client.auth_app_id('foo', 'bar')

        assert self.client.token == result['auth']['client_token']
        assert (await self.client.is_authenticated())

        self.client.token = self.root_token()
        await self.client.disable_auth_backend('app-id')

    async def test_create_app_id(self):
        if 'app-id/' not in (await self.client.list_auth_backends()):
            await self.client.enable_auth_backend('app-id')

        await self.client.create_app_id('testappid', policies='not_root', display_name='displayname')

        result = await self.client.read('auth/app-id/map/app-id/testappid')
        lib_result = await self.client.get_app_id('testappid')
        del result['request_id']
        del lib_result['request_id']
        assert result == lib_result

        assert result['data']['key'] == 'testappid'
        assert result['data']['display_name'] == 'displayname'
        assert result['data']['value'] == 'not_root'
        await self.client.delete_app_id('testappid')
        assert (await self.client.get_app_id('testappid'))['data'] is None

        self.client.token = self.root_token()
        await self.client.disable_auth_backend('app-id')

    async def test_cubbyhole_auth(self):
        orig_token = self.client.token

        resp = await self.client.create_token(lease='6h', wrap_ttl='1h')
        assert resp['wrap_info']['ttl'] == 3600

        wrapped_token = resp['wrap_info']['token']
        await self.client.auth_cubbyhole(wrapped_token)
        assert self.client.token != orig_token
        assert self.client.token != wrapped_token
        assert (await self.client.is_authenticated())

        self.client.token = orig_token
        assert (await self.client.is_authenticated())

    async def test_create_user_id(self):
        if 'app-id/' not in (await self.client.list_auth_backends()):
            await self.client.enable_auth_backend('app-id')

        await self.client.create_app_id('testappid', policies='not_root', display_name='displayname')
        await self.client.create_user_id('testuserid', app_id='testappid')

        result = await self.client.read('auth/app-id/map/user-id/testuserid')
        lib_result = await self.client.get_user_id('testuserid')
        del result['request_id']
        del lib_result['request_id']
        assert result == lib_result

        assert result['data']['key'] == 'testuserid'
        assert result['data']['value'] == 'testappid'

        result = await self.client.auth_app_id('testappid', 'testuserid')

        assert self.client.token == result['auth']['client_token']
        assert (await self.client.is_authenticated())
        self.client.token = self.root_token()
        await self.client.delete_user_id('testuserid')
        assert (await self.client.get_user_id('testuserid'))['data'] is None

        self.client.token = self.root_token()
        await self.client.disable_auth_backend('app-id')

    async def test_create_role(self):
        if 'approle/' in (await self.client.list_auth_backends()):
            await self.client.disable_auth_backend('approle')
        await self.client.enable_auth_backend('approle')

        await self.client.create_role('testrole')

        result = await self.client.read('auth/approle/role/testrole')
        lib_result = await self.client.get_role('testrole')
        del result['request_id']
        del lib_result['request_id']

        assert result == lib_result
        self.client.token = self.root_token()
        await self.client.disable_auth_backend('approle')

    async def test_delete_role(self):
        test_role_name = 'test-role'
        if 'approle/' in (await self.client.list_auth_backends()):
            await self.client.disable_auth_backend('approle')
        await self.client.enable_auth_backend('approle')

        await self.client.create_role(test_role_name)
        # We add a second dummy test role so we can still hit the /role?list=true route after deleting the first role
        await self.client.create_role('test-role-2')

        # Ensure our created role shows up when calling list_roles as expected
        result = await self.client.list_roles()
        actual_list_role_keys = result['data']['keys']
        self.assertIn(
            member=test_role_name,
            container=actual_list_role_keys,
        )

        # Now delete the role and verify its absence when calling list_roles
        await self.client.delete_role(test_role_name)
        result = await self.client.list_roles()
        actual_list_role_keys = result['data']['keys']
        self.assertNotIn(
            member=test_role_name,
            container=actual_list_role_keys,
        )

        # reset test environment
        self.client.token = self.root_token()
        await self.client.disable_auth_backend('approle')

    async def test_create_delete_role_secret_id(self):
        if 'approle/' in (await self.client.list_auth_backends()):
            await self.client.disable_auth_backend('approle')
        await self.client.enable_auth_backend('approle')

        await self.client.create_role('testrole')
        create_result = await self.client.create_role_secret_id('testrole', {'foo': 'bar'})
        secret_id = create_result['data']['secret_id']
        result = await self.client.get_role_secret_id('testrole', secret_id)
        assert result['data']['metadata']['foo'] == 'bar'
        await self.client.delete_role_secret_id('testrole', secret_id)
        assert (await self.client.get_role_secret_id('testrole', secret_id)) is None
        self.client.token = self.root_token()
        await self.client.disable_auth_backend('approle')

    async def test_auth_approle(self):
        if 'approle/' in (await self.client.list_auth_backends()):
            await self.client.disable_auth_backend('approle')
        await self.client.enable_auth_backend('approle')

        await self.client.create_role('testrole')
        create_result = await self.client.create_role_secret_id('testrole', {'foo': 'bar'})
        secret_id = create_result['data']['secret_id']
        role_id = await self.client.get_role_id('testrole')
        result = await self.client.auth_approle(role_id, secret_id)
        assert result['auth']['metadata']['foo'] == 'bar'
        assert self.client.token == result['auth']['client_token']
        assert await self.client.is_authenticated()
        self.client.token = self.root_token()
        await self.client.disable_auth_backend('approle')

    async def test_auth_approle_dont_use_token(self):
        if 'approle/' in (await self.client.list_auth_backends()):
            await self.client.disable_auth_backend('approle')
        await self.client.enable_auth_backend('approle')

        await self.client.create_role('testrole')
        create_result = await self.client.create_role_secret_id('testrole', {'foo': 'bar'})
        secret_id = create_result['data']['secret_id']
        role_id = await self.client.get_role_id('testrole')
        result = await self.client.auth_approle(role_id, secret_id, use_token=False)
        assert result['auth']['metadata']['foo'] == 'bar'
        assert self.client.token != result['auth']['client_token']
        self.client.token = self.root_token()
        await self.client.disable_auth_backend('approle')

    async def test_transit_read_write(self):
        if 'transit/' in (await self.client.list_secret_backends()):
            await self.client.disable_secret_backend('transit')
        await self.client.enable_secret_backend('transit')

        await self.client.transit_create_key('foo')
        result = await self.client.transit_read_key('foo')
        assert not result['data']['exportable']

        await self.client.transit_create_key('foo_export', exportable=True, key_type="ed25519")
        result = await self.client.transit_read_key('foo_export')
        assert result['data']['exportable']
        assert result['data']['type'] == 'ed25519'

        await self.client.enable_secret_backend('transit', mount_point='bar')
        await self.client.transit_create_key('foo', mount_point='bar')
        result = await self.client.transit_read_key('foo', mount_point='bar')
        assert not result['data']['exportable']

    async def test_transit_list_keys(self):
        if 'transit/' in (await self.client.list_secret_backends()):
            await self.client.disable_secret_backend('transit')
        await self.client.enable_secret_backend('transit')

        await self.client.transit_create_key('foo1')
        await self.client.transit_create_key('foo2')
        await self.client.transit_create_key('foo3')

        result = await self.client.transit_list_keys()
        assert result['data']['keys'] == ["foo1", "foo2", "foo3"]

    async def test_transit_update_delete_keys(self):
        if 'transit/' in (await self.client.list_secret_backends()):
            await self.client.disable_secret_backend('transit')
        await self.client.enable_secret_backend('transit')

        await self.client.transit_create_key('foo')
        await self.client.transit_update_key('foo', deletion_allowed=True)
        result = await self.client.transit_read_key('foo')
        assert result['data']['deletion_allowed']

        await self.client.transit_delete_key('foo')

        try:
            await self.client.transit_read_key('foo')
        except exceptions.InvalidPath:
            assert True
        else:
            assert False

    async def test_transit_rotate_key(self):
        if 'transit/' in (await self.client.list_secret_backends()):
            await self.client.disable_secret_backend('transit')
        await self.client.enable_secret_backend('transit')

        await self.client.transit_create_key('foo')

        await self.client.transit_rotate_key('foo')
        response = await self.client.transit_read_key('foo')
        assert '2' in response['data']['keys']

        await self.client.transit_rotate_key('foo')
        response = await self.client.transit_read_key('foo')
        assert '3' in response['data']['keys']

    async def test_transit_export_key(self):
        if 'transit/' in (await self.client.list_secret_backends()):
            await self.client.disable_secret_backend('transit')
        await self.client.enable_secret_backend('transit')

        await self.client.transit_create_key('foo', exportable=True)
        response = await self.client.transit_export_key('foo', key_type='encryption-key')
        assert response is not None

    async def test_transit_encrypt_data(self):
        if 'transit/' in (await self.client.list_secret_backends()):
            await self.client.disable_secret_backend('transit')
        await self.client.enable_secret_backend('transit')

        await self.client.transit_create_key('foo')
        ciphertext_resp = (await self.client.transit_encrypt_data('foo', 'abbaabba'))['data']['ciphertext']
        plaintext_resp = (await self.client.transit_decrypt_data('foo', ciphertext_resp))['data']['plaintext']
        assert plaintext_resp == 'abbaabba'

    async def test_transit_rewrap_data(self):
        if 'transit/' in (await self.client.list_secret_backends()):
            await self.client.disable_secret_backend('transit')
        await self.client.enable_secret_backend('transit')

        await self.client.transit_create_key('foo')
        ciphertext_resp = (await self.client.transit_encrypt_data('foo', 'abbaabba'))['data']['ciphertext']

        await self.client.transit_rotate_key('foo')
        response_wrap = (await self.client.transit_rewrap_data('foo', ciphertext=ciphertext_resp))['data']['ciphertext']
        plaintext_resp = (await self.client.transit_decrypt_data('foo', response_wrap))['data']['plaintext']
        assert plaintext_resp == 'abbaabba'

    async def test_transit_generate_data_key(self):
        if 'transit/' in (await self.client.list_secret_backends()):
            await self.client.disable_secret_backend('transit')
        await self.client.enable_secret_backend('transit')

        await self.client.transit_create_key('foo')

        response_plaintext = (await self.client.transit_generate_data_key('foo', key_type='plaintext'))['data']['plaintext']
        assert response_plaintext

        response_ciphertext = (await self.client.transit_generate_data_key('foo', key_type='wrapped'))['data']
        assert 'ciphertext' in response_ciphertext
        assert 'plaintext' not in response_ciphertext

    async def test_transit_generate_rand_bytes(self):
        if 'transit/' in (await self.client.list_secret_backends()):
            await self.client.disable_secret_backend('transit')
        await self.client.enable_secret_backend('transit')

        response_data = (await self.client.transit_generate_rand_bytes(data_bytes=4))['data']['random_bytes']
        assert response_data

    async def test_transit_hash_data(self):
        if 'transit/' in (await self.client.list_secret_backends()):
            await self.client.disable_secret_backend('transit')
        await self.client.enable_secret_backend('transit')

        response_hash = (await self.client.transit_hash_data('abbaabba'))['data']['sum']
        assert len(response_hash) == 64

        response_hash = (await self.client.transit_hash_data('abbaabba', algorithm="sha2-512"))['data']['sum']
        assert len(response_hash) == 128

    async def test_transit_generate_verify_hmac(self):
        if 'transit/' in (await self.client.list_secret_backends()):
            await self.client.disable_secret_backend('transit')
        await self.client.enable_secret_backend('transit')

        await self.client.transit_create_key('foo')

        response_hmac = (await self.client.transit_generate_hmac('foo', 'abbaabba'))['data']['hmac']
        assert response_hmac
        verify_resp = (await self.client.transit_verify_signed_data('foo', 'abbaabba', hmac=response_hmac))['data']['valid']
        assert verify_resp

        response_hmac = (await self.client.transit_generate_hmac('foo', 'abbaabba', algorithm='sha2-512'))['data']['hmac']
        assert response_hmac
        verify_resp = (await self.client.transit_verify_signed_data('foo', 'abbaabba',
                                                                    algorithm='sha2-512', hmac=response_hmac))['data']['valid']
        assert verify_resp

    async def test_transit_sign_verify_signature_data(self):
        if 'transit/' in (await self.client.list_secret_backends()):
            await self.client.disable_secret_backend('transit')
        await self.client.enable_secret_backend('transit')

        await self.client.transit_create_key('foo', key_type='ed25519')

        signed_resp = (await self.client.transit_sign_data('foo', 'abbaabba'))['data']['signature']
        assert signed_resp
        verify_resp = (await self.client.transit_verify_signed_data('foo', 'abbaabba', signature=signed_resp))['data']['valid']
        assert verify_resp

        signed_resp = (await self.client.transit_sign_data('foo', 'abbaabba', algorithm='sha2-512'))['data']['signature']
        assert signed_resp
        verify_resp = (await self.client.transit_verify_signed_data('foo', 'abbaabba',
                                                                    algorithm='sha2-512',
                                                                    signature=signed_resp))['data']['valid']
        assert verify_resp

    async def test_missing_token(self):
        client = create_client()
        assert not (await client.is_authenticated())
        await client.close()

    async def test_invalid_token(self):
        client = create_client(token='not-a-real-token')
        assert not (await client.is_authenticated())
        await client.close()

    async def test_illegal_token(self):
        client = create_client(token='token-with-new-line\n')
        try:
            await client.is_authenticated()
        except ValueError as e:
            assert 'Invalid header value' in str(e)
        await client.close()

    async def test_broken_token(self):
        client = create_client(token='\x1b')
        try:
            await client.is_authenticated()
        except exceptions.InvalidRequest as e:
            assert "invalid header value" in str(e)
        await client.close()

    async def test_client_authenticated(self):
        assert (await self.client.is_authenticated())

    async def test_client_logout(self):
        self.client.logout()
        assert not (await self.client.is_authenticated())

    async def test_revoke_self_token(self):
        if 'userpass/' in (await self.client.list_auth_backends()):
            await self.client.disable_auth_backend('userpass')

        await self.client.enable_auth_backend('userpass')

        await self.client.write('auth/userpass/users/testuser', password='testpass', policies='not_root')

        await self.client.auth_userpass('testuser', 'testpass')

        await self.client.revoke_self_token()
        assert not (await self.client.is_authenticated())

    async def test_rekey_multi(self):
        cls = type(self)

        assert not (await self.client.rekey_status)['started']

        await self.client.start_rekey()
        assert (await self.client.rekey_status)['started']

        await self.client.cancel_rekey()
        assert not (await self.client.rekey_status)['started']

        result = await self.client.start_rekey()

        keys = cls.manager.keys

        result = await self.client.rekey_multi(keys, nonce=result['nonce'])
        assert result['complete']

        cls.manager.keys = result['keys']
        cls.manager.unseal()

    async def test_rotate(self):
        status = await self.client.key_status

        await self.client.rotate()

        assert (await self.client.key_status)['term'] > status['term']

    async def test_tls_auth(self):
        await self.client.enable_auth_backend('cert')

        with open('test/client-cert.pem') as fp:
            certificate = fp.read()

        await self.client.write('auth/cert/certs/test', display_name='test',
                                policies='not_root', certificate=certificate)

        await self.client.auth_tls()

    async def test_gh51(self):
        key = 'secret/http://test.com'

        await self.client.write(key, foo='bar')

        result = await self.client.read(key)

        assert result['data']['foo'] == 'bar'

    async def test_token_accessor(self):
        # Create token, check accessor is provided
        result = await self.client.create_token(lease='1h')
        token_accessor = result['auth'].get('accessor', None)
        assert token_accessor

        # Look up token by accessor, make sure token is excluded from results
        lookup = await self.client.lookup_token(token_accessor, accessor=True)
        assert lookup['data']['accessor'] == token_accessor
        assert not lookup['data']['id']

        # Revoke token using the accessor
        await self.client.revoke_token(token_accessor, accessor=True)

        # Look up by accessor should fail
        with self.assertRaises(exceptions.InvalidRequest):
            lookup = await self.client.lookup_token(token_accessor, accessor=True)

        # As should regular lookup
        with self.assertRaises(exceptions.Forbidden):
            lookup = await self.client.lookup_token(result['auth']['client_token'])

    async def test_wrapped_token_success(self):
        wrap = await self.client.create_token(wrap_ttl='1m')

        # Unwrap token
        result = await self.client.unwrap(wrap['wrap_info']['token'])
        assert result['auth']['client_token']

        # Validate token
        lookup = await self.client.lookup_token(result['auth']['client_token'])
        assert result['auth']['client_token'] == lookup['data']['id']

    async def test_wrapped_token_intercept(self):
        wrap = await self.client.create_token(wrap_ttl='1m')

        # Intercept wrapped token
        await self.client.unwrap(wrap['wrap_info']['token'])

        # Attempt to retrieve the token after it's been intercepted
        with self.assertRaises(exceptions.InvalidRequest):
            await self.client.unwrap(wrap['wrap_info']['token'])

    async def test_wrapped_token_cleanup(self):
        wrap = await self.client.create_token(wrap_ttl='1m')

        _token = self.client.token
        await self.client.unwrap(wrap['wrap_info']['token'])
        assert self.client.token == _token

    async def test_wrapped_token_revoke(self):
        wrap = await self.client.create_token(wrap_ttl='1m')

        # Revoke token before it's unwrapped
        await self.client.revoke_token(wrap['wrap_info']['wrapped_accessor'], accessor=True)

        # Unwrap token anyway
        result = await self.client.unwrap(wrap['wrap_info']['token'])
        assert result['auth']['client_token']

        # Attempt to validate token
        with self.assertRaises(exceptions.Forbidden):
            await self.client.lookup_token(result['auth']['client_token'])

    async def test_wrapped_client_token_success(self):
        wrap = await self.client.create_token(wrap_ttl='1m')
        self.client.token = wrap['wrap_info']['token']

        # Unwrap token
        result = await self.client.unwrap()
        assert result['auth']['client_token']

        # Validate token
        self.client.token = result['auth']['client_token']
        lookup = await self.client.lookup_token(result['auth']['client_token'])
        assert result['auth']['client_token'] == lookup['data']['id']

    async def test_wrapped_client_token_intercept(self):
        wrap = await self.client.create_token(wrap_ttl='1m')
        self.client.token = wrap['wrap_info']['token']

        # Intercept wrapped token
        await self.client.unwrap()

        # Attempt to retrieve the token after it's been intercepted
        with self.assertRaises(exceptions.InvalidRequest):
            await self.client.unwrap()

    async def test_wrapped_client_token_cleanup(self):
        wrap = await self.client.create_token(wrap_ttl='1m')

        _token = self.client.token
        self.client.token = wrap['wrap_info']['token']
        await self.client.unwrap()

        assert self.client.token != wrap
        assert self.client.token != _token

    async def test_wrapped_client_token_revoke(self):
        wrap = await self.client.create_token(wrap_ttl='1m')

        # Revoke token before it's unwrapped
        await self.client.revoke_token(wrap['wrap_info']['wrapped_accessor'], accessor=True)

        # Unwrap token anyway
        self.client.token = wrap['wrap_info']['token']
        result = await self.client.unwrap()
        assert result['auth']['client_token']

        # Attempt to validate token
        with self.assertRaises(exceptions.Forbidden):
            await self.client.lookup_token(result['auth']['client_token'])

    async def test_create_token_explicit_max_ttl(self):

        token = await self.client.create_token(ttl='30m', explicit_max_ttl='5m')

        assert token['auth']['client_token']

        assert token['auth']['lease_duration'] == 300

        # Validate token
        lookup = await self.client.lookup_token(token['auth']['client_token'])
        assert token['auth']['client_token'] == lookup['data']['id']

    async def test_create_token_max_ttl(self):

        token = await self.client.create_token(ttl='5m')

        assert token['auth']['client_token']

        assert token['auth']['lease_duration'] == 300

        # Validate token
        lookup = await self.client.lookup_token(token['auth']['client_token'])
        assert token['auth']['client_token'] == lookup['data']['id']

    async def test_create_token_periodic(self):

        token = await self.client.create_token(period='30m')

        assert token['auth']['client_token']

        assert token['auth']['lease_duration'] == 1800

        # Validate token
        lookup = await self.client.lookup_token(token['auth']['client_token'])
        assert token['auth']['client_token'] == lookup['data']['id']
        assert lookup['data']['period'] == 1800

    async def test_token_roles(self):
        # No roles, list_token_roles == None
        before = await self.client.list_token_roles()
        assert not before

        # Create token role
        assert (await self.client.create_token_role('testrole')).status == 204

        # List token roles
        during = (await self.client.list_token_roles())['data']['keys']
        assert len(during) == 1
        assert during[0] == 'testrole'

        # Delete token role
        await self.client.delete_token_role('testrole')

        # No roles, list_token_roles == None
        after = await self.client.list_token_roles()
        assert not after

    async def test_create_token_w_role(self):
        # Create policy
        await self.prep_policy('testpolicy')

        # Create token role w/ policy
        assert (await self.client.create_token_role(
            'testrole', allowed_policies='testpolicy')).status == 204

        # Create token against role
        token = await self.client.create_token(lease='1h', role='testrole')
        assert token['auth']['client_token']
        assert token['auth']['policies'] == ['default', 'testpolicy']

        # Cleanup
        await self.client.delete_token_role('testrole')
        await self.client.delete_policy('testpolicy')

    async def test_ec2_role_crud(self):
        if 'aws-ec2/' in (await self.client.list_auth_backends()):
            await self.client.disable_auth_backend('aws-ec2')
        await self.client.enable_auth_backend('aws-ec2')

        # create a policy to associate with the role
        await self.prep_policy('ec2rolepolicy')

        # attempt to get a list of roles before any exist
        no_roles = await self.client.list_ec2_roles()
        # doing so should succeed and return None
        assert (no_roles is None)

        # test binding by AMI ID (the old way, to ensure backward compatibility)
        await self.client.create_ec2_role('foo',
                                          'ami-notarealami',
                                          policies='ec2rolepolicy')

        # test binding by Account ID
        await self.client.create_ec2_role('bar',
                                          bound_account_id='123456789012',
                                          policies='ec2rolepolicy')

        # test binding by IAM Role ARN
        await self.client.create_ec2_role('baz',
                                          bound_iam_role_arn='arn:aws:iam::123456789012:role/mockec2role',
                                          policies='ec2rolepolicy')

        # test binding by instance profile ARN
        await self.client.create_ec2_role('qux',
                                          bound_iam_instance_profile_arn='arn:aws:iam::123456789012:instance-profile/mockprofile',
                                          policies='ec2rolepolicy')

        # test binding by bound region
        await self.client.create_ec2_role('quux',
                                          bound_region='ap-northeast-2',
                                          policies='ec2rolepolicy')

        # test binding by bound vpc id
        await self.client.create_ec2_role('corge',
                                          bound_vpc_id='vpc-1a123456',
                                          policies='ec2rolepolicy')

        # test binding by bound subnet id
        await self.client.create_ec2_role('grault',
                                          bound_subnet_id='subnet-123a456',
                                          policies='ec2rolepolicy')

        roles = await self.client.list_ec2_roles()

        assert('foo' in roles['data']['keys'])
        assert('bar' in roles['data']['keys'])
        assert('baz' in roles['data']['keys'])
        assert('qux' in roles['data']['keys'])
        assert('quux' in roles['data']['keys'])
        assert('corge' in roles['data']['keys'])
        assert('grault' in roles['data']['keys'])

        foo_role = await self.client.get_ec2_role('foo')
        assert ('ami-notarealami' in foo_role['data']['bound_ami_id'])
        assert ('ec2rolepolicy' in foo_role['data']['policies'])

        bar_role = await self.client.get_ec2_role('bar')
        assert ('123456789012' in bar_role['data']['bound_account_id'])
        assert ('ec2rolepolicy' in bar_role['data']['policies'])

        baz_role = await self.client.get_ec2_role('baz')
        assert ('arn:aws:iam::123456789012:role/mockec2role' in baz_role['data']['bound_iam_role_arn'])
        assert ('ec2rolepolicy' in baz_role['data']['policies'])

        qux_role = await self.client.get_ec2_role('qux')
        assert('arn:aws:iam::123456789012:instance-profile/mockprofile' in qux_role['data']['bound_iam_instance_profile_arn'])
        assert('ec2rolepolicy' in qux_role['data']['policies'])

        quux_role = await self.client.get_ec2_role('quux')
        assert('ap-northeast-2' in quux_role['data']['bound_region'])
        assert('ec2rolepolicy' in quux_role['data']['policies'])

        corge_role = await self.client.get_ec2_role('corge')
        assert('vpc-1a123456' in corge_role['data']['bound_vpc_id'])
        assert('ec2rolepolicy' in corge_role['data']['policies'])

        grault_role = await self.client.get_ec2_role('grault')
        assert('subnet-123a456' in grault_role['data']['bound_subnet_id'])
        assert('ec2rolepolicy' in grault_role['data']['policies'])

        # teardown
        await self.client.delete_ec2_role('foo')
        await self.client.delete_ec2_role('bar')
        await self.client.delete_ec2_role('baz')
        await self.client.delete_ec2_role('qux')
        await self.client.delete_ec2_role('quux')
        await self.client.delete_ec2_role('corge')
        await self.client.delete_ec2_role('grault')

        await self.client.delete_policy('ec2rolepolicy')

        await self.client.disable_auth_backend('aws-ec2')

    async def test_ec2_role_token_lifespan(self):
        if 'aws-ec2/' not in (await self.client.list_auth_backends()):
            await self.client.enable_auth_backend('aws-ec2')

        # create a policy to associate with the role
        await self.prep_policy('ec2rolepolicy')

        # create a role with no TTL
        await self.client.create_ec2_role('foo',
                                          'ami-notarealami',
                                          policies='ec2rolepolicy')

        # create a role with a 1hr TTL
        await self.client.create_ec2_role('bar',
                                          'ami-notarealami',
                                          ttl='1h',
                                          policies='ec2rolepolicy')

        # create a role with a 3-day max TTL
        await self.client.create_ec2_role('baz',
                                          'ami-notarealami',
                                          max_ttl='72h',
                                          policies='ec2rolepolicy')

        # create a role with 1-day period
        await self.client.create_ec2_role('qux',
                                          'ami-notarealami',
                                          period='24h',
                                          policies='ec2rolepolicy')

        foo_role = await self.client.get_ec2_role('foo')
        assert (foo_role['data']['ttl'] == 0)

        bar_role = await self.client.get_ec2_role('bar')
        assert (bar_role['data']['ttl'] == 3600)

        baz_role = await self.client.get_ec2_role('baz')
        assert (baz_role['data']['max_ttl'] == 259200)

        qux_role = await self.client.get_ec2_role('qux')
        assert (qux_role['data']['period'] == 86400)

        # teardown
        await self.client.delete_ec2_role('foo')
        await self.client.delete_ec2_role('bar')
        await self.client.delete_ec2_role('baz')
        await self.client.delete_ec2_role('qux')

        await self.client.delete_policy('ec2rolepolicy')

        await self.client.disable_auth_backend('aws-ec2')

    async def test_start_generate_root_with_completion(self):
        test_otp = 'RSMGkAqBH5WnVLrDTbZ+UQ=='

        self.assertFalse((await self.client.generate_root_status)['started'])
        response = await self.client.start_generate_root(
            key=test_otp,
            otp=True,
        )
        self.assertTrue((await self.client.generate_root_status)['started'])

        nonce = response['nonce']
        for key in self.manager.keys[0:3]:
            response = await self.client.generate_root(
                key=key,
                nonce=nonce,
            )
        self.assertFalse((await self.client.generate_root_status)['started'])

        # Decode the token provided in the last response. Root token decoding logic derived from:
        # https://github.com/hashicorp/vault/blob/284600fbefc32d8ab71b6b9d1d226f2f83b56b1d/command/operator_generate_root.go#L289
        b64decoded_root_token = b64decode(response['encoded_root_token'])
        if sys.version_info > (3, 0):
            # b64decoding + bytes XOR'ing to decode the new root token in python 3.x
            int_encoded_token = int.from_bytes(b64decoded_root_token, sys.byteorder)
            int_otp = int.from_bytes(b64decode(test_otp), sys.byteorder)
            xord_otp_and_token = int_otp ^ int_encoded_token
            token_hex_string = xord_otp_and_token.to_bytes(len(b64decoded_root_token), sys.byteorder).hex()
        else:
            # b64decoding + bytes XOR'ing to decode the new root token in python 2.7
            otp_and_token = zip(b64decode(test_otp), b64decoded_root_token)
            xord_otp_and_token = ''.join(chr(ord(y) ^ ord(x)) for (x, y) in otp_and_token)
            token_hex_string = binascii.hexlify(xord_otp_and_token)

        new_root_token = str(UUID(token_hex_string))

        # Assert our new root token is properly formed and authenticated
        self.client.token = new_root_token
        if await self.client.is_authenticated():
            self.root_token = new_root_token
        else:
            # If our new token was unable to authenticate, set the test client's token back to the original value
            self.client.token = self.root_token
            self.fail('Unable to authenticate with the newly generated root token.')

    async def test_start_generate_root_then_cancel(self):
        test_otp = 'RSMGkAqBH5WnVLrDTbZ+UQ=='

        self.assertFalse((await self.client.generate_root_status)['started'])
        await self.client.start_generate_root(
            key=test_otp,
            otp=True,
        )
        self.assertTrue((await self.client.generate_root_status)['started'])

        await self.client.cancel_generate_root()
        self.assertFalse((await self.client.generate_root_status)['started'])

    async def test_auth_ec2_alternate_mount_point_with_no_client_token_exception(self):
        test_mount_point = 'aws-custom-path'
        # Turn on the aws-ec2 backend with a custom mount_point path specified.
        if '{0}/'.format(test_mount_point) in (await self.client.list_auth_backends()):
            await self.client.disable_auth_backend(test_mount_point)
        await self.client.enable_auth_backend('aws-ec2', mount_point=test_mount_point)

        # Drop the client's token to replicate a typical end user's use of any auth method.
        # I.e., its reasonable to expect the method is being called to _retrieve_ a token in the first place.
        self.client.token = None

        # Load a mock PKCS7 encoded self-signed certificate to stand in for a real document from the AWS identity service.
        with open('test/identity_document.p7b') as fp:
            pkcs7 = fp.read()

        # When attempting to auth (POST) to an auth backend mounted at a different path than the default, we expect a
        # generic 'missing client token' response from Vault.
        with self.assertRaises(exceptions.InvalidRequest) as assertRaisesContext:
            await self.client.auth_ec2(pkcs7=pkcs7)

        expected_exception_message = 'missing client token'
        actual_exception_message = str(assertRaisesContext.exception)
        self.assertEqual(expected_exception_message, actual_exception_message)

        # Reset test state.
        self.client.token = self.root_token()
        await self.client.disable_auth_backend(mount_point=test_mount_point)

    async def test_auth_ec2_alternate_mount_point_with_no_client_token(self):
        test_mount_point = 'aws-custom-path'
        # Turn on the aws-ec2 backend with a custom mount_point path specified.
        if '{0}/'.format(test_mount_point) in (await self.client.list_auth_backends()):
            await self.client.disable_auth_backend(test_mount_point)
        await self.client.enable_auth_backend('aws-ec2', mount_point=test_mount_point)

        # Drop the client's token to replicate a typical end user's use of any auth method.
        # I.e., its reasonable to expect the method is being called to _retrieve_ a token in the first place.
        self.client.token = None

        # Load a mock PKCS7 encoded self-signed certificate to stand in for a real document from the AWS identity service.
        with open('test/identity_document.p7b') as fp:
            pkcs7 = fp.read()

        # If our custom path is respected, we'll still end up with Vault's inability to decrypt our dummy PKCS7 string.
        # However this exception indicates we're correctly hitting the expected auth endpoint.
        with self.assertRaises(exceptions.InternalServerError) as assertRaisesContext:
            await self.client.auth_ec2(pkcs7=pkcs7, mount_point=test_mount_point)

        expected_exception_message = 'failed to decode the PEM encoded PKCS#7 signature'
        actual_exception_message = str(assertRaisesContext.exception)
        self.assertEqual(expected_exception_message, actual_exception_message)

        # Reset test state.
        self.client.token = self.root_token()
        await self.client.disable_auth_backend(mount_point=test_mount_point)

    async def test_auth_gcp_alternate_mount_point_with_no_client_token_exception(self):
        test_mount_point = 'gcp-custom-path'
        # Turn on the gcp backend with a custom mount_point path specified.
        if '{0}/'.format(test_mount_point) in (await self.client.list_auth_backends()):
            await self.client.disable_auth_backend(test_mount_point)
        await self.client.enable_auth_backend('gcp', mount_point=test_mount_point)

        # Drop the client's token to replicate a typical end user's use of any auth method.
        # I.e., its reasonable to expect the method is being called to _retrieve_ a token in the first place.
        self.client.token = None

        # Load a mock JWT stand in for a real document from GCP.
        with open('test/example.jwt') as fp:
            jwt = fp.read()

        # When attempting to auth (POST) to an auth backend mounted at a different path than the default, we expect a
        # generic 'missing client token' response from Vault.
        with self.assertRaises(exceptions.InvalidRequest) as assertRaisesContext:
            await self.client.auth_gcp('example-role', jwt)

        expected_exception_message = 'missing client token'
        actual_exception_message = str(assertRaisesContext.exception)
        self.assertEqual(expected_exception_message, actual_exception_message)

        # Reset test state.
        self.client.token = self.root_token()
        await self.client.disable_auth_backend(mount_point=test_mount_point)

    async def test_tune_auth_backend(self):
        test_backend_type = 'approle'
        test_mount_point = 'tune-approle'
        test_description = 'this is a test auth backend'
        test_max_lease_ttl = 12345678
        if '{0}/'.format(test_mount_point) in (await self.client.list_auth_backends()):
            await self.client.disable_auth_backend(test_mount_point)
        await self.client.enable_auth_backend(
            backend_type='approle',
            mount_point=test_mount_point
        )

        expected_status_code = 204
        response = await self.client.tune_auth_backend(
            backend_type=test_backend_type,
            mount_point=test_mount_point,
            description=test_description,
            max_lease_ttl=test_max_lease_ttl,
        )
        self.assertEqual(
            first=expected_status_code,
            second=response.status,
        )

        response = await self.client.get_auth_backend_tuning(
            backend_type=test_backend_type,
            mount_point=test_mount_point
        )

        self.assertEqual(
            first=test_max_lease_ttl,
            second=response['data']['max_lease_ttl']
        )

        await self.client.disable_auth_backend(mount_point=test_mount_point)

    async def test_kv2_secret_backend(self):
        if 'test/' in (await self.client.list_secret_backends()):
            await self.client.disable_secret_backend('test')
        await self.client.enable_secret_backend('kv', mount_point='test', options={'version': '2'})

        secret_backends = await self.client.list_secret_backends()

        assert 'test/' in secret_backends
        self.assertDictEqual(secret_backends['test/']['options'], {'version': '2'})

        await self.client.disable_secret_backend('test')

    async def test_create_kubernetes_configuration(self):
        expected_status_code = 204
        test_mount_point = 'k8s'

        # Turn on the kubernetes backend with a custom mount_point path specified.
        if '{0}/'.format(test_mount_point) in (await self.client.list_auth_backends()):
            await self.client.disable_auth_backend(test_mount_point)
        await self.client.enable_auth_backend('kubernetes', mount_point=test_mount_point)

        with open('test/client-cert.pem') as fp:
            certificate = fp.read()
            response = await self.client.create_kubernetes_configuration(
                kubernetes_host='127.0.0.1:80',
                pem_keys=[certificate],
                mount_point=test_mount_point,
            )
        self.assertEqual(
            first=expected_status_code,
            second=response.status,
        )

        # Reset integration test state
        await self.client.disable_auth_backend(mount_point=test_mount_point)

    async def test_get_kubernetes_configuration(self):
        test_host = '127.0.0.1:80'
        test_mount_point = 'k8s'

        # Turn on the kubernetes backend with a custom mount_point path specified.
        if '{0}/'.format(test_mount_point) in (await self.client.list_auth_backends()):
            await self.client.disable_auth_backend(test_mount_point)
        await self.client.enable_auth_backend('kubernetes', mount_point=test_mount_point)
        with open('test/client-cert.pem') as fp:
            certificate = fp.read()
            await self.client.create_kubernetes_configuration(
                kubernetes_host=test_host,
                pem_keys=[certificate],
                mount_point=test_mount_point,
            )

        # Test that we can retrieve the configuration
        response = await self.client.get_kubernetes_configuration(
            mount_point=test_mount_point
        )
        self.assertIn(
            member='data',
            container=response,
        )
        self.assertEqual(
            first=test_host,
            second=response['data'].get('kubernetes_host')
        )

        # Reset integration test state
        await self.client.disable_auth_backend(mount_point=test_mount_point)

    async def test_create_kubernetes_role(self):
        test_role_name = 'test_role'
        test_mount_point = 'k8s'
        expected_status_code = 204

        # Turn on the kubernetes backend with a custom mount_point path specified.
        if '{0}/'.format(test_mount_point) in (await self.client.list_auth_backends()):
            await self.client.disable_auth_backend(test_mount_point)
        await self.client.enable_auth_backend('kubernetes', mount_point=test_mount_point)

        with open('test/client-cert.pem') as fp:
            certificate = fp.read()
            await self.client.create_kubernetes_configuration(
                kubernetes_host='127.0.0.1:80',
                pem_keys=[certificate],
                mount_point=test_mount_point,
            )

        # Test that we can createa role
        response = await self.client.create_kubernetes_role(
            name=test_role_name,
            bound_service_account_names='*',
            bound_service_account_namespaces='vault_test',
            mount_point=test_mount_point,
        )
        self.assertEqual(
            first=expected_status_code,
            second=response.status,
        )

        # Reset integration test state
        await self.client.disable_auth_backend(mount_point=test_mount_point)

    async def test_get_kubernetes_role(self):
        test_role_name = 'test_role'
        test_mount_point = 'k8s'
        test_bound_service_account_namespaces = ['vault-test']

        # Turn on the kubernetes backend with a custom mount_point path specified.
        if '{0}/'.format(test_mount_point) in (await self.client.list_auth_backends()):
            await self.client.disable_auth_backend(test_mount_point)
        await self.client.enable_auth_backend('kubernetes', mount_point=test_mount_point)

        with open('test/client-cert.pem') as fp:
            certificate = fp.read()
            await self.client.create_kubernetes_configuration(
                kubernetes_host='127.0.0.1:80',
                pem_keys=[certificate],
                mount_point=test_mount_point,
            )

        # Test that we can createa role
        await self.client.create_kubernetes_role(
            name=test_role_name,
            bound_service_account_names='*',
            bound_service_account_namespaces=test_bound_service_account_namespaces,
            mount_point=test_mount_point,
        )
        response = await self.client.get_kubernetes_role(
            name=test_role_name,
            mount_point=test_mount_point,
        )
        self.assertIn(
            member='data',
            container=response,
        )
        self.assertEqual(
            first=test_bound_service_account_namespaces,
            second=response['data'].get('bound_service_account_namespaces')
        )
        # Reset integration test state
        await self.client.disable_auth_backend(mount_point=test_mount_point)

    async def test_list_kubernetes_roles(self):
        test_role_name = 'test_role'
        test_mount_point = 'k8s'
        test_bound_service_account_namespaces = ['vault-test']

        # Turn on the kubernetes backend with a custom mount_point path specified.
        if '{0}/'.format(test_mount_point) in (await self.client.list_auth_backends()):
            await self.client.disable_auth_backend(test_mount_point)
        await self.client.enable_auth_backend('kubernetes', mount_point=test_mount_point)

        with open('test/client-cert.pem') as fp:
            certificate = fp.read()
            await self.client.create_kubernetes_configuration(
                kubernetes_host='127.0.0.1:80',
                pem_keys=[certificate],
                mount_point=test_mount_point,
            )

        # Test that we can createa role
        await self.client.create_kubernetes_role(
            name=test_role_name,
            bound_service_account_names='*',
            bound_service_account_namespaces=test_bound_service_account_namespaces,
            mount_point=test_mount_point,
        )
        response = await self.client.list_kubernetes_roles(
            mount_point=test_mount_point,
        )
        self.assertIn(
            member='data',
            container=response,
        )
        self.assertEqual(
            first=[test_role_name],
            second=response['data'].get('keys')
        )
        # Reset integration test state
        await self.client.disable_auth_backend(mount_point=test_mount_point)

    async def test_delete_kubernetes_role(self):
        test_role_name = 'test_role'
        test_mount_point = 'k8s'
        expected_status_code = 204

        # Turn on the kubernetes backend with a custom mount_point path specified.
        if '{0}/'.format(test_mount_point) in (await self.client.list_auth_backends()):
            await self.client.disable_auth_backend(test_mount_point)
        await self.client.enable_auth_backend('kubernetes', mount_point=test_mount_point)

        with open('test/client-cert.pem') as fp:
            certificate = fp.read()
            await self.client.create_kubernetes_configuration(
                kubernetes_host='127.0.0.1:80',
                pem_keys=[certificate],
                mount_point=test_mount_point,
            )

        await self.client.create_kubernetes_role(
            name=test_role_name,
            bound_service_account_names='*',
            bound_service_account_namespaces='vault_test',
            mount_point=test_mount_point,
        )
        # Test that we can delete a role
        response = await self.client.delete_kubernetes_role(
            role=test_role_name,
            mount_point=test_mount_point,
        )
        self.assertEqual(
            first=expected_status_code,
            second=response.status,
        )

        # Reset integration test state
        await self.client.disable_auth_backend(mount_point=test_mount_point)

    async def test_auth_kubernetes(self):
        test_role_name = 'test_role'
        test_host = '127.0.0.1:80'
        test_mount_point = 'k8s'

        # Turn on the kubernetes backend with a custom mount_point path specified.
        if '{0}/'.format(test_mount_point) in (await self.client.list_auth_backends()):
            await self.client.disable_auth_backend(test_mount_point)
        await self.client.enable_auth_backend('kubernetes', mount_point=test_mount_point)
        with open('test/client-cert.pem') as fp:
            certificate = fp.read()
            await self.client.create_kubernetes_configuration(
                kubernetes_host=test_host,
                pem_keys=[certificate],
                mount_point=test_mount_point,
            )

        await self.client.create_kubernetes_role(
            name=test_role_name,
            bound_service_account_names='*',
            bound_service_account_namespaces='vault_test',
            mount_point=test_mount_point,
        )

        # Test that we can authenticate
        with open('test/example.jwt') as fp:
            test_jwt = fp.read()
            with self.assertRaises(exceptions.InternalServerError) as assertRaisesContext:
                # we don't actually have a valid JWT to provide, so this method will throw an exception
                await self.client.auth_kubernetes(
                    role=test_role_name,
                    jwt=test_jwt,
                    mount_point=test_mount_point,
                )

        expected_exception_message = 'claim "iss" is invalid'
        actual_exception_message = str(assertRaisesContext.exception)
        self.assertEqual(expected_exception_message, actual_exception_message)

        # Reset integration test state
        await self.client.disable_auth_backend(mount_point=test_mount_point)



import json
from base64 import b64decode
from datetime import datetime
import asynctest

from async_hvac import AsyncClient


class TestAwsIamMethods(asynctest.TestCase):
    """Unit tests providing coverage for AWS (EC2) auth backend-related methods/routes."""

    @asynctest.mock.patch('async_hvac.aws_utils.datetime')
    @asynctest.mock.patch('async_hvac.v1.AsyncClient.auth')
    async def test_auth_aws_iam(self, auth_mock, datetime_mock):
        datetime_mock.utcnow.return_value = datetime(2015, 8, 30, 12, 36, 0)
        client = AsyncClient()
        await client.auth_aws_iam('AKIDEXAMPLE', 'wJalrXUtnFEMI/K7MDENG+bPxRfiCYEXAMPLEKEY')

        auth_mock.assert_called()
        args, kwargs = auth_mock.call_args
        actual_params = kwargs['json']

        actual_iam_http_request_method = actual_params['iam_http_request_method']
        self.assertEqual('POST', actual_iam_http_request_method)

        actual_iam_request_url = b64decode(actual_params['iam_request_url']).decode('utf-8')
        self.assertEqual('https://sts.amazonaws.com/', actual_iam_request_url)

        expected_auth_header_parts = [
            'Credential=AKIDEXAMPLE/20150830/us-east-1/sts/aws4_request',
            'SignedHeaders=content-length;content-type;host;x-amz-date',
            'Signature=0268ea4a725deae1116f5228d6b177fb047f9f3a9e1c5fd4baa0dc1fbb0d1a99',
        ]
        expected_iam_request_headers = {
            'Authorization': ['{0} {1}'.format('AWS4-HMAC-SHA256', ', '.join(expected_auth_header_parts))],
            'Content-Length': ['43'],
            'Content-Type': ['application/x-www-form-urlencoded; charset=utf-8'],
            'Host': ['sts.amazonaws.com'],
            'X-Amz-Date': ['20150830T123600Z'],
        }
        actual_iam_request_headers = json.loads(b64decode(actual_params['iam_request_headers']))
        self.assertEqual(expected_iam_request_headers, actual_iam_request_headers)

        actual_iam_request_body = b64decode(actual_params['iam_request_body']).decode('utf-8')
        self.assertEqual('Action=GetCallerIdentity&Version=2011-06-15', actual_iam_request_body)

        actual_role = actual_params['role']
        self.assertEqual('', actual_role)
        await client.close()



from asynctest import TestCase

from parameterized import parameterized

from async_hvac import AsyncClient
from async_hvac.tests.util import RequestsMocker


class TestKubernetesMethods(TestCase):
    """Unit tests providing coverage for Kubernetes auth backend-related methods/routes."""

    @parameterized.expand([
        ("default mount point", None, '127.0.0.1:80', ['test_key']),
        ("custom mount point", "k8s", 'some_k8s_host.com', ['test_key']),
    ])
    @RequestsMocker()
    async def test_create_kubernetes_configuration(self, test_label, mount_point, kubernetes_host, pem_keys, requests_mocker):
        expected_status_code = 204
        mock_url = 'http://127.0.0.1:8200/v1/auth/{0}/config'.format(
            'kubernetes' if mount_point is None else mount_point,
        )
        requests_mocker.register_uri(
            method='POST',
            url=mock_url,
            status_code=expected_status_code,
        )
        client = AsyncClient()

        test_arguments = dict(
            kubernetes_host=kubernetes_host,
            pem_keys=pem_keys,
        )
        if mount_point:
            test_arguments['mount_point'] = mount_point

        actual_response = await client.create_kubernetes_configuration(**test_arguments)

        self.assertEqual(
            first=expected_status_code,
            second=actual_response.status,
        )
        await client.close()

    @parameterized.expand([
        ("default mount point", None),
        ("custom mount point", "k8s"),
    ])
    @RequestsMocker()
    async def test_get_kubernetes_configuration(self, test_label, mount_point, requests_mocker):
        expected_status_code = 200
        mock_response = {
            'auth': None,
            'data': {
                'kubernetes_ca_cert': '',
                'kubernetes_host': '127.0.0.1:80',
                'pem_keys': ['some key'],
                'token_reviewer_jwt': ''
            },
            'lease_duration': 0,
            'lease_id': '',
            'renewable': False,
            'request_id': '12687b5f-b4f5-2ba4-aae2-2a8d7e53ca55',
            'warnings': None,
            'wrap_info': None
        }
        mock_url = 'http://127.0.0.1:8200/v1/auth/{0}/config'.format(
            'kubernetes' if mount_point is None else mount_point,
        )
        requests_mocker.register_uri(
            method='GET',
            url=mock_url,
            status_code=expected_status_code,
            json=mock_response,
        )
        client = AsyncClient()

        test_arguments = dict()
        if mount_point:
            test_arguments['mount_point'] = mount_point

        actual_response = await client.get_kubernetes_configuration(**test_arguments)

        self.assertEqual(
            first=mock_response,
            second=actual_response,
        )
        await client.close()

    @parameterized.expand([
        ("default mount point", None, "application1", '*', 'some-namespace'),
        ("custom mount point", "k8s", "application2", 'some-service-account', '*'),
    ])
    @RequestsMocker()
    async def test_create_role(self, test_label, mount_point, role_name, bound_service_account_names, bound_service_account_namespaces, requests_mocker):
        expected_status_code = 204
        mock_url = 'http://127.0.0.1:8200/v1/auth/{0}/role/{1}'.format(
            'kubernetes' if mount_point is None else mount_point,
            role_name,
        )
        requests_mocker.register_uri(
            method='POST',
            url=mock_url,
            status_code=expected_status_code,
        )
        client = AsyncClient()

        test_arguments = dict(
            name=role_name,
            bound_service_account_names=bound_service_account_names,
            bound_service_account_namespaces=bound_service_account_namespaces,
        )
        if mount_point:
            test_arguments['mount_point'] = mount_point
        actual_response = await client.create_kubernetes_role(**test_arguments)

        self.assertEqual(
            first=expected_status_code,
            second=actual_response.status,
        )
        await client.close()

    @parameterized.expand([
        ("default mount point", None, "application1"),
        ("custom mount point", "k8s", "application2"),
    ])
    @RequestsMocker()
    async def test_get_role(self, test_label, mount_point, role_name, requests_mocker):
        expected_status_code = 200
        mock_response = {
            "auth": None,
            "data": {
                "bind_secret_id": True,
                "bound_cidr_list": "",
                "period": 0,
                "policies": [
                    "default"
                ],
                "secret_id_num_uses": 0,
                "secret_id_ttl": 0,
                "token_max_ttl": 900,
                "token_num_uses": 0,
                "token_ttl": 600
            },
            "lease_duration": 0,
            "lease_id": "",
            "renewable": False,
            "request_id": "0aab655f-ecd2-b3d4-3817-35b5bdfd3f28",
            "warnings": None,
            "wrap_info": None
        }
        mock_url = 'http://127.0.0.1:8200/v1/auth/{0}/role/{1}'.format(
            'kubernetes' if mount_point is None else mount_point,
            role_name,
        )
        requests_mocker.register_uri(
            method='GET',
            url=mock_url,
            status_code=expected_status_code,
            json=mock_response,
        )
        client = AsyncClient()

        test_arguments = dict(
            name=role_name,
        )
        if mount_point:
            test_arguments['mount_point'] = mount_point

        actual_response = await client.get_kubernetes_role(**test_arguments)

        self.assertEqual(
            first=mock_response,
            second=actual_response,
        )
        await client.close()

    @parameterized.expand([
        ("default mount point", None, ['test-role-1', 'test-role-2']),
        ("custom mount point", "k8s", ['test-role']),
    ])
    @RequestsMocker()
    async def test_list_kubernetes_roles(self, test_label, mount_point, role_names, requests_mocker):
        expected_status_code = 200
        mock_response = {
            "auth": None,
            "data": {
                "keys": role_names,
            },
            "lease_duration": 0,
            "lease_id": "",
            "renewable": False,
            "request_id": "e4c219fb-0a78-2be2-8d3c-b3715dccb920",
            "warnings": None,
            "wrap_info": None
        }
        mock_url = 'http://127.0.0.1:8200/v1/auth/{0}/role?list=true'.format(
            'kubernetes' if mount_point is None else mount_point,
        )
        requests_mocker.register_uri(
            method='GET',
            url=mock_url,
            status_code=expected_status_code,
            json=mock_response,
        )
        client = AsyncClient()

        test_arguments = dict()
        if mount_point:
            test_arguments['mount_point'] = mount_point
        actual_response = await client.list_kubernetes_roles(**test_arguments)

        # ensure we received our mock response data back successfully
        self.assertEqual(mock_response, actual_response)
        await client.close()

    @parameterized.expand([
        ("default mount point", None, "application1"),
        ("custom mount point", "k8s", "application2"),
    ])
    @RequestsMocker()
    async def test_delete_kubernetes_role(self, test_label, mount_point, role_name, requests_mocker):
        expected_status_code = 204
        mock_url = 'http://127.0.0.1:8200/v1/auth/{0}/role/{1}'.format(
            'kubernetes' if mount_point is None else mount_point,
            role_name,
        )
        requests_mocker.register_uri(
            method='DELETE',
            url=mock_url,
            status_code=expected_status_code,
        )
        client = AsyncClient()

        test_arguments = dict(
            role=role_name,
        )
        if mount_point:
            test_arguments['mount_point'] = mount_point

        actual_response = await client.delete_kubernetes_role(**test_arguments)

        self.assertEqual(
            first=expected_status_code,
            second=actual_response.status,
        )
        await client.close()

    @parameterized.expand([
        ("default mount point", "custom_role", "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9", None),
        ("custom mount point", "custom_role", "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9", "gcp-not-default")
    ])
    @RequestsMocker()
    async def test_auth_kubernetes(self, test_label, test_role, test_jwt, mount_point, requests_mocker):
        mock_response = {
            'auth': {
                'accessor': 'accessor-1234-5678-9012-345678901234',
                'client_token': 'cltoken-1234-5678-9012-345678901234',
                'lease_duration': 10000,
                'metadata': {
                    'role': 'custom_role',
                    'service_account_name': 'vault-auth',
                    'service_account_namespace': 'default',
                    'service_account_secret_name': 'vault-auth-token-pd21c',
                    'service_account_uid': 'aa9aa8ff-98d0-11e7-9bb7-0800276d99bf'
                },
                'policies': [
                    'default',
                    'custom_role'
                ],
                'renewable': True
            },
            'data': None,
            'lease_duration': 0,
            'lease_id': '',
            'renewable': False,
            'request_id': 'requesti-1234-5678-9012-345678901234',
            'warnings': [],
            'wrap_info': None
        }
        mock_url = 'http://127.0.0.1:8200/v1/auth/{0}/login'.format(
                'kubernetes' if mount_point is None else mount_point)
        requests_mocker.register_uri(
            method='POST',
            url=mock_url,
            json=mock_response
        )
        client = AsyncClient()
        test_arguments = dict(
            role=test_role,
            jwt=test_jwt,
        )
        if mount_point:
            test_arguments['mount_point'] = mount_point
        actual_response = await client.auth_kubernetes(**test_arguments)

        # ensure we received our mock response data back successfully
        self.assertEqual(mock_response, actual_response)
        await client.close()



from asynctest import TestCase

from parameterized import parameterized

from async_hvac import AsyncClient
from async_hvac.tests.util import RequestsMocker


class TestAwsEc2Methods(TestCase):
    """Unit tests providing coverage for AWS (EC2) auth backend-related methods/routes."""

    @parameterized.expand([
        ("default mount point", None),
        ("custom mount point", 'aws-ec2'),
    ])
    @RequestsMocker()
    async def test_auth_ec2(self, test_label, mount_point, requests_mocker):
        mock_response = {
            'auth': {
                'accessor': 'accessor-1234-5678-9012-345678901234',
                'client_token': 'cltoken-1234-5678-9012-345678901234',
                'lease_duration': 10000,
                'metadata': {
                    'account_id': '12345678912',
                    'ami_id': 'ami-someami',
                    'instance_id': 'i-instanceid',
                    'nonce': 'thenonce-1234-5678-9012-345678901234',
                    'region': 'us-east-1',
                    'role': 'custom_role',
                    'role_tag_max_ttl': '0s'
                },
                'policies': [
                    'default',
                    'custom_role'
                ],
                'renewable': True
            },
            'data': None,
            'lease_duration': 0,
            'lease_id': '',
            'renewable': False,
            'request_id': 'requesti-1234-5678-9012-345678901234',
            'warnings': [],
            'wrap_info': None
        }
        mock_url = 'http://127.0.0.1:8200/v1/auth/{0}/login'.format(
            'aws-ec2' if mount_point is None else mount_point
        )
        requests_mocker.register_uri(
            method='POST',
            url=mock_url,
            json=mock_response
        )
        client = AsyncClient()

        if mount_point is None:
            actual_response = await client.auth_ec2(
                pkcs7='mock_pcks7'
            )
        else:
            actual_response = await client.auth_ec2(
                pkcs7='mock_pcks7',
                mount_point=mount_point
            )

        # ensure we received our mock response data back successfully
        self.assertEqual(mock_response, actual_response)
        await client.close()

    @parameterized.expand([
        ("default mount point", None),
        ("custom mount point", 'aws-ec2'),
    ])
    @RequestsMocker()
    async def test_create_vault_ec2_client_configuration(self, test_label, mount_point, requests_mocker):
        test_access_key = 'AKIAABCDEFGUE1234567'
        test_secret_key = 'thisisasecretyall'
        expected_status_code = 204
        mock_url = 'http://127.0.0.1:8200/v1/auth/{0}/config/client'.format(
            'aws-ec2' if mount_point is None else mount_point
        )
        requests_mocker.register_uri(
            method='POST',
            url=mock_url,
            status_code=expected_status_code,
        )
        client = AsyncClient()

        if mount_point is None:
            actual_response = await client.create_vault_ec2_client_configuration(
                access_key=test_access_key,
                secret_key=test_secret_key,
            )
        else:
            actual_response = await client.create_vault_ec2_client_configuration(
                access_key=test_access_key,
                secret_key=test_secret_key,
                mount_point=mount_point
            )

        self.assertEqual(
            first=expected_status_code,
            second=actual_response.status
        )
        await client.close()

    @parameterized.expand([
        ("default mount point", None),
        ("custom mount point", 'aws-ec2'),
    ])
    @RequestsMocker()
    async def test_get_vault_ec2_client_configuration(self, test_label, mount_point, requests_mocker):
        mock_response = {
          "data": {
            "secret_key": "vCtSM8ZUEQ3mOFVlYPBQkf2sO6F/W7a5TVzrl3Oj",
            "access_key": "VKIAJBRHKH6EVTTNXDHA",
            "endpoint": "",
            "iam_endpoint": "",
            "sts_endpoint": "",
            "iam_server_id_header_value": ""
          }
        }
        expected_status_code = 200
        mock_url = 'http://127.0.0.1:8200/v1/auth/{0}/config/client'.format(
            'aws-ec2' if mount_point is None else mount_point
        )
        requests_mocker.register_uri(
            method='GET',
            url=mock_url,
            json=mock_response,
            status_code=expected_status_code,
        )
        client = AsyncClient()

        if mount_point is None:
            actual_response = await client.get_vault_ec2_client_configuration()
        else:
            actual_response = await client.get_vault_ec2_client_configuration(
                mount_point=mount_point
            )

        self.assertEqual(
            first=mock_response,
            second=actual_response,
        )
        await client.close()

    @parameterized.expand([
        ("default mount point", None),
        ("custom mount point", 'aws-ec2'),
    ])
    @RequestsMocker()
    async def test_delete_vault_ec2_client_configuration(self, test_label, mount_point, requests_mocker):
        expected_status_code = 204
        mock_url = 'http://127.0.0.1:8200/v1/auth/{0}/config/client'.format(
            'aws-ec2' if mount_point is None else mount_point
        )
        requests_mocker.register_uri(
            method='DELETE',
            url=mock_url,
            status_code=expected_status_code,
        )
        client = AsyncClient()

        if mount_point is None:
            actual_response = await client.delete_vault_ec2_client_configuration()
        else:
            actual_response = await client.delete_vault_ec2_client_configuration(
                mount_point=mount_point
            )

        self.assertEqual(
            first=expected_status_code,
            second=actual_response.status,
        )
        await client.close()

    @parameterized.expand([
        ("default mount point", None, 'my-cool-cert-1'),
        ("custom mount point", 'aws-ec2', 'my-cool-cert-2'),
    ])
    @RequestsMocker()
    async def test_create_vault_ec2_certificate_configuration(self, test_label, mount_point, cert_name, requests_mocker):
        test_cert_info = 'this is some test cert info'
        expected_status_code = 204
        mock_url = 'http://127.0.0.1:8200/v1/auth/{0}/config/certificate/{1}'.format(
            'aws-ec2' if mount_point is None else mount_point,
            cert_name,
        )
        requests_mocker.register_uri(
            method='POST',
            url=mock_url,
            status_code=expected_status_code,
        )
        client = AsyncClient()

        if mount_point is None:
            actual_response = await client.create_vault_ec2_certificate_configuration(
                cert_name=cert_name,
                aws_public_cert=test_cert_info,
            )
        else:
            actual_response = await client.create_vault_ec2_certificate_configuration(
                cert_name=cert_name,
                aws_public_cert=test_cert_info,
                mount_point=mount_point
            )

        self.assertEqual(
            first=expected_status_code,
            second=actual_response.status,
        )
        await client.close()

    @parameterized.expand([
        ("default mount point", None, 'my-cool-cert-1'),
        ("custom mount point", 'aws-ec2', 'my-cool-cert-2'),
    ])
    @RequestsMocker()
    async def test_get_vault_ec2_certificate_configuration(self, test_label, mount_point, cert_name, requests_mocker):
        mock_response = {
            "data": {
                "aws_public_cert": "-----BEGIN CERTIFICATE-----\nblah blah9K\n-----END CERTIFICATE-----\n",
                "type": "pkcs7"
            }
        }
        expected_status_code = 200
        mock_url = 'http://127.0.0.1:8200/v1/auth/{0}/config/certificate/{1}'.format(
            'aws-ec2' if mount_point is None else mount_point,
            cert_name,
        )
        requests_mocker.register_uri(
            method='GET',
            url=mock_url,
            json=mock_response,
            status_code=expected_status_code,
        )
        client = AsyncClient()

        if mount_point is None:
            actual_response = await client.get_vault_ec2_certificate_configuration(
                cert_name=cert_name,
            )
        else:
            actual_response = await client.get_vault_ec2_certificate_configuration(
                cert_name=cert_name,
                mount_point=mount_point
            )

        self.assertEqual(
            first=mock_response,
            second=actual_response,
        )
        await client.close()

    @parameterized.expand([
        ("default mount point", None),
        ("custom mount point", 'aws-ec2'),
    ])
    @RequestsMocker()
    async def test_list_vault_ec2_certificate_configurations(self, test_label, mount_point, requests_mocker):
        mock_response = {
            "data": {
                "keys": [
                    "cert1"
                ]
            }
        }

        expected_status_code = 200
        mock_url = 'http://127.0.0.1:8200/v1/auth/{0}/config/certificates?list=true'.format(
            'aws-ec2' if mount_point is None else mount_point,
        )
        requests_mocker.register_uri(
            method='GET',
            url=mock_url,
            json=mock_response,
            status_code=expected_status_code,
        )
        client = AsyncClient()

        if mount_point is None:
            actual_response = await client.list_vault_ec2_certificate_configurations()
        else:
            actual_response = await client.list_vault_ec2_certificate_configurations(
                mount_point=mount_point
            )

        self.assertEqual(
            first=mock_response,
            second=actual_response,
        )
        await client.close()

    @parameterized.expand([
        ("default mount point", None, 'my-role-1'),
        ("custom mount point", 'aws-ec2', 'my-role-2'),
    ])
    @RequestsMocker()
    async def test_create_ec2_role(self, test_label, mount_point, role_name, requests_mocker):
        expected_status_code = 204
        mock_url = 'http://127.0.0.1:8200/v1/auth/{0}/role/{1}'.format(
            'aws-ec2' if mount_point is None else mount_point,
            role_name,
        )
        requests_mocker.register_uri(
            method='POST',
            url=mock_url,
            status_code=expected_status_code,
        )
        client = AsyncClient()

        if mount_point is None:
            actual_response = await client.create_ec2_role(
                role=role_name
            )
        else:
            actual_response = await client.create_ec2_role(
                role=role_name,
                mount_point=mount_point
            )

        self.assertEqual(
            first=expected_status_code,
            second=actual_response.status,
        )
        await client.close()

    @parameterized.expand([
        ("default mount point", None, 'my-role-1'),
        ("custom mount point", 'aws-ec2', 'my-role-2'),
    ])
    @RequestsMocker()
    async def test_get_ec2_role(self, test_label, mount_point, role_name, requests_mocker):
        mock_response = {
            "data": {
                "bound_ami_id": ["ami-fce36987"],
                "role_tag": "",
                "policies": [
                    "default",
                    "dev",
                    "prod"
                ],
                "max_ttl": 1800000,
                "disallow_reauthentication": False,
                "allow_instance_migration": False
            }
        }
        expected_status_code = 200
        mock_url = 'http://127.0.0.1:8200/v1/auth/{0}/role/{1}'.format(
            'aws-ec2' if mount_point is None else mount_point,
            role_name,
        )
        requests_mocker.register_uri(
            method='GET',
            url=mock_url,
            json=mock_response,
            status_code=expected_status_code,
        )
        client = AsyncClient()

        if mount_point is None:
            actual_response = await client.get_ec2_role(
                role=role_name
            )
        else:
            actual_response = await client.get_ec2_role(
                role=role_name,
                mount_point=mount_point
            )

        self.assertEqual(
            first=mock_response,
            second=actual_response,
        )
        await client.close()

    @parameterized.expand([
        ("default mount point", None, 'my-role-1'),
        ("custom mount point", 'aws-ec2', 'my-role-2'),
    ])
    @RequestsMocker()
    async def test_delete_ec2_role(self, test_label, mount_point, role_name, requests_mocker):
        expected_status_code = 204
        mock_url = 'http://127.0.0.1:8200/v1/auth/{0}/role/{1}'.format(
            'aws-ec2' if mount_point is None else mount_point,
            role_name,
        )
        requests_mocker.register_uri(
            method='DELETE',
            url=mock_url,
            status_code=expected_status_code,
        )
        client = AsyncClient()

        if mount_point is None:
            actual_response = await client.delete_ec2_role(
                role=role_name
            )
        else:
            actual_response = await client.delete_ec2_role(
                role=role_name,
                mount_point=mount_point
            )

        self.assertEqual(
            first=expected_status_code,
            second=actual_response.status,
        )
        await client.close()

    @parameterized.expand([
        ("default mount point", None),
        ("custom mount point", 'aws-ec2'),
    ])
    @RequestsMocker()
    async def test_list_ec2_roles(self, test_label, mount_point, requests_mocker):
        mock_response = {
            "data": {
                "keys": [
                    "dev-role",
                    "prod-role"
                ]
            }
        }
        expected_status_code = 200
        mock_url = 'http://127.0.0.1:8200/v1/auth/{0}/roles?list=true'.format(
            'aws-ec2' if mount_point is None else mount_point,
        )
        requests_mocker.register_uri(
            method='GET',
            url=mock_url,
            json=mock_response,
            status_code=expected_status_code,
        )
        client = AsyncClient()

        if mount_point is None:
            actual_response = await client.list_ec2_roles()
        else:
            actual_response = await client.list_ec2_roles(
                mount_point=mount_point
            )

        self.assertEqual(
            first=mock_response,
            second=actual_response,
        )
        await client.close()

    @parameterized.expand([
        ("default mount point", None, 'my-role-1'),
        ("custom mount point", 'aws-ec2', 'my-role-2'),
    ])
    @RequestsMocker()
    async def test_create_ec2_role_tag(self, test_label, mount_point, role_name, requests_mocker):
        mock_response = {
            "data": {
                "tag_value": "v1:09Vp0qGuyB8=:r=dev-role:p=default,dev-api:d=false:t=300h0m0s:uPLKCQxqsefRhrp1qmVa1wsQVUXXJG8UZP/pJIdVyOI=",
                "tag_key": "VaultRole"
            }
        }
        expected_status_code = 200
        mock_url = 'http://127.0.0.1:8200/v1/auth/{0}/role/{1}/tag'.format(
            'aws-ec2' if mount_point is None else mount_point,
            role_name
        )
        requests_mocker.register_uri(
            method='POST',
            url=mock_url,
            json=mock_response,
            status_code=expected_status_code,
        )
        client = AsyncClient()

        if mount_point is None:
            actual_response = await client.create_ec2_role_tag(
                role=role_name,
            )
        else:
            actual_response = await client.create_ec2_role_tag(
                role=role_name,
                mount_point=mount_point
            )

        self.assertEqual(
            first=mock_response,
            second=await actual_response.json(),
        )
        await client.close()



from __future__ import unicode_literals

import json
import ssl
from base64 import b64encode

import aiohttp
from async_hvac import aws_utils, exceptions

try:
    import hcl

    has_hcl_parser = True
except ImportError:
    has_hcl_parser = False


class AsyncClient(object):
    def __init__(self, url='http://127.0.0.1:8200', token=None,
                 cert=None, verify=True, timeout=30, proxies=None,
                 allow_redirects=True, session=None, loop=None):

        self.allow_redirects = allow_redirects
        self._session = session
        self.token = token

        self._url = url
        self._timeout = timeout
        self._cert = cert
        self._proxies = proxies
        self._loop = loop
        self._sslcontext = None
        if verify and self._cert:
            self._sslcontext = ssl.create_default_context(cafile=verify)
            self._sslcontext.load_cert_chain(self._cert[0], self._cert[1])
        else:
            self._sslcontext = False

    def __enter__(self):
        raise TypeError("Use async with instead")

    def __exit__(self, exc_type, exc_val, exc_tb):
        # __exit__ should exist in pair with __enter__ but never executed
        pass  # pragma: no cover

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    @property
    def session(self):
        if not self._session:
            self._session = aiohttp.ClientSession(
                loop=self._loop,
                timeout=aiohttp.ClientTimeout(total=self._timeout)
            )
        return self._session

    async def read(self, path, wrap_ttl=None):
        """
        GET /<path>
        """
        try:
            return await (await self._get('/v1/{0}'.format(path), wrap_ttl=wrap_ttl)).json()
        except exceptions.InvalidPath:
            return None

    async def list(self, path):
        """
        GET /<path>?list=true
        """
        try:
            payload = {
                'list': 'true'
            }
            return await (await self._get('/v1/{}'.format(path), params=payload)).json()
        except exceptions.InvalidPath:
            return None

    async def write(self, path, wrap_ttl=None, **kwargs):
        """
        POST /<path>
        """
        response = await self._post('/v1/{0}'.format(path), json=kwargs, wrap_ttl=wrap_ttl)

        if response.status == 200:
            return await response.json()

    def delete(self, path):
        """
        DELETE /<path>
        """
        return self._delete('/v1/{0}'.format(path))

    async def unwrap(self, token=None):
        """
        POST /sys/wrapping/unwrap
        X-Vault-Token: <token>
        """
        if token:
            payload = {
                'token': token
            }
            return await (await self._post('/v1/sys/wrapping/unwrap', json=payload)).json()
        else:
            return await (await self._post('/v1/sys/wrapping/unwrap')).json()

    async def is_initialized(self):
        """
        GET /sys/init
        """
        return (await (await self._get('/v1/sys/init')).json())['initialized']

    async def initialize(self, secret_shares=5, secret_threshold=3, pgp_keys=None):
        """
        PUT /sys/init
        """
        params = {
            'secret_shares': secret_shares,
            'secret_threshold': secret_threshold,
        }

        if pgp_keys:
            if len(pgp_keys) != secret_shares:
                raise ValueError('Length of pgp_keys must equal secret shares')

            params['pgp_keys'] = pgp_keys

        return await (await self._put('/v1/sys/init', json=params)).json()

    @property
    async def seal_status(self):
        """
        GET /sys/seal-status
        """
        return await (await self._get('/v1/sys/seal-status')).json()

    async def is_sealed(self):
        return (await self.seal_status)['sealed']

    def seal(self):
        """
        PUT /sys/seal
        """
        return self._put('/v1/sys/seal')

    async def unseal_reset(self):
        """
        PUT /sys/unseal
        """
        params = {
            'reset': True,
        }
        return await (await self._put('/v1/sys/unseal', json=params)).json()

    async def unseal(self, key):
        """
        PUT /sys/unseal
        """
        params = {
            'key': key,
        }

        return await (await self._put('/v1/sys/unseal', json=params)).json()

    async def unseal_multi(self, keys):
        result = None

        for key in keys:
            result = await self.unseal(key)
            if not result['sealed']:
                break

        return result

    @property
    async def generate_root_status(self):
        """
        GET /sys/generate-root/attempt
        """
        return await (await self._get('/v1/sys/generate-root/attempt')).json()

    async def start_generate_root(self, key, otp=False):
        """
        PUT /sys/generate-root/attempt
        """
        params = {}
        if otp:
            params['otp'] = key
        else:
            params['pgp_key'] = key

        return await (await self._put('/v1/sys/generate-root/attempt', json=params)).json()

    async def generate_root(self, key, nonce):
        """
        PUT /sys/generate-root/update
        """
        params = {
            'key': key,
            'nonce': nonce,
        }

        return await (await self._put('/v1/sys/generate-root/update', json=params)).json()

    async def cancel_generate_root(self):
        """
        DELETE /sys/generate-root/attempt
        """

        return (await self._delete('/v1/sys/generate-root/attempt')).status == 204

    @property
    async def key_status(self):
        """
        GET /sys/key-status
        """
        return await (await self._get('/v1/sys/key-status')).json()

    def rotate(self):
        """
        PUT /sys/rotate
        """
        return self._put('/v1/sys/rotate')

    @property
    async def rekey_status(self):
        """
        GET /sys/rekey/init
        """
        return await (await self._get('/v1/sys/rekey/init')).json()

    async def start_rekey(self, secret_shares=5, secret_threshold=3, pgp_keys=None,
                          backup=False):
        """
        PUT /sys/rekey/init
        """
        params = {
            'secret_shares': secret_shares,
            'secret_threshold': secret_threshold,
        }

        if pgp_keys:
            if len(pgp_keys) != secret_shares:
                raise ValueError('Length of pgp_keys must equal secret shares')

            params['pgp_keys'] = pgp_keys
            params['backup'] = backup

        resp = await self._put('/v1/sys/rekey/init', json=params)
        if resp.text:
            return await resp.json()

    def cancel_rekey(self):
        """
        DELETE /sys/rekey/init
        """
        return self._delete('/v1/sys/rekey/init')

    async def rekey(self, key, nonce=None):
        """
        PUT /sys/rekey/update
        """
        params = {
            'key': key,
        }

        if nonce:
            params['nonce'] = nonce

        return await (await self._put('/v1/sys/rekey/update', json=params)).json()

    async def rekey_multi(self, keys, nonce=None):
        result = None

        for key in keys:
            result = await self.rekey(key, nonce=nonce)
            if result.get('complete'):
                break

        return result

    async def get_backed_up_keys(self):
        """
        GET /sys/rekey/backup
        """
        return await (await self._get('/v1/sys/rekey/backup')).json()

    @property
    async def ha_status(self):
        """
        GET /sys/leader
        """
        return await (await self._get('/v1/sys/leader')).json()

    async def renew_secret(self, lease_id, increment=None):
        """
        PUT /sys/leases/renew
        """
        params = {
            'lease_id': lease_id,
            'increment': increment,
        }
        return await (await self._post('/v1/sys/renew', json=params)).json()

    def revoke_secret(self, lease_id):
        """
        PUT /sys/revoke/<lease id>
        """
        return self._put('/v1/sys/revoke/{0}'.format(lease_id))

    def revoke_secret_prefix(self, path_prefix):
        """
        PUT /sys/revoke-prefix/<path prefix>
        """
        return self._put('/v1/sys/revoke-prefix/{0}'.format(path_prefix))

    def revoke_self_token(self):
        """
        PUT /auth/token/revoke-self
        """
        return self._put('/v1/auth/token/revoke-self')

    async def list_secret_backends(self):
        """
        GET /sys/mounts
        """
        return await (await self._get('/v1/sys/mounts')).json()

    def enable_secret_backend(self, backend_type, description=None, mount_point=None, config=None, options=None):
        """
        POST /sys/auth/<mount point>
        """
        if not mount_point:
            mount_point = backend_type

        params = {
            'type': backend_type,
            'description': description,
            'config': config,
            'options': options,
        }

        return self._post('/v1/sys/mounts/{0}'.format(mount_point), json=params)

    def tune_secret_backend(self, backend_type, mount_point=None, default_lease_ttl=None, max_lease_ttl=None):
        """
        POST /sys/mounts/<mount point>/tune
        """

        if not mount_point:
            mount_point = backend_type

        params = {
            'default_lease_ttl': default_lease_ttl,
            'max_lease_ttl': max_lease_ttl
        }

        return self._post('/v1/sys/mounts/{0}/tune'.format(mount_point), json=params)

    async def get_secret_backend_tuning(self, backend_type, mount_point=None):
        """
        GET /sys/mounts/<mount point>/tune
        """
        if not mount_point:
            mount_point = backend_type

        return await (await self._get('/v1/sys/mounts/{0}/tune'.format(mount_point))).json()

    def disable_secret_backend(self, mount_point):
        """
        DELETE /sys/mounts/<mount point>
        """
        return self._delete('/v1/sys/mounts/{0}'.format(mount_point))

    def remount_secret_backend(self, from_mount_point, to_mount_point):
        """
        POST /sys/remount
        """
        params = {
            'from': from_mount_point,
            'to': to_mount_point,
        }

        return self._post('/v1/sys/remount', json=params)

    async def list_policies(self):
        """
        GET /sys/policy
        """
        return (await (await self._get('/v1/sys/policy')).json())['policies']

    async def get_policy(self, name, parse=False):
        """
        GET /sys/policy/<name>
        """
        try:
            policy = (await (await self._get('/v1/sys/policy/{0}'.format(name))).json())['rules']
            if parse:
                if not has_hcl_parser:
                    raise ImportError('pyhcl is required for policy parsing')

                policy = hcl.loads(policy)

            return policy
        except exceptions.InvalidPath:
            return None

    def set_policy(self, name, rules):
        """
        PUT /sys/policy/<name>
        """

        if isinstance(rules, dict):
            rules = json.dumps(rules)

        params = {
            'rules': rules,
        }

        return self._put('/v1/sys/policy/{0}'.format(name), json=params)

    def delete_policy(self, name):
        """
        DELETE /sys/policy/<name>
        """
        return self._delete('/v1/sys/policy/{0}'.format(name))

    async def list_audit_backends(self):
        """
        GET /sys/audit
        """
        return await (await self._get('/v1/sys/audit')).json()

    def enable_audit_backend(self, backend_type, description=None, options=None, name=None):
        """
        POST /sys/audit/<name>
        """
        if not name:
            name = backend_type

        params = {
            'type': backend_type,
            'description': description,
            'options': options,
        }

        return self._post('/v1/sys/audit/{0}'.format(name), json=params)

    def disable_audit_backend(self, name):
        """
        DELETE /sys/audit/<name>
        """
        return self._delete('/v1/sys/audit/{0}'.format(name))

    async def audit_hash(self, name, input):
        """
        POST /sys/audit-hash
        """
        params = {
            'input': input,
        }
        return await (await self._post('/v1/sys/audit-hash/{0}'.format(name), json=params)).json()

    async def create_token(self, role=None, token_id=None, policies=None, meta=None,
                           no_parent=False, lease=None, display_name=None,
                           num_uses=None, no_default_policy=False,
                           ttl=None, orphan=False, wrap_ttl=None, renewable=None,
                           explicit_max_ttl=None, period=None):
        """
        POST /auth/token/create
        POST /auth/token/create/<role>
        POST /auth/token/create-orphan
        """
        params = {
            'id': token_id,
            'policies': policies,
            'meta': meta,
            'no_parent': no_parent,
            'display_name': display_name,
            'num_uses': num_uses,
            'no_default_policy': no_default_policy,
            'renewable': renewable
        }

        if lease:
            params['lease'] = lease
        else:
            params['ttl'] = ttl
            params['explicit_max_ttl'] = explicit_max_ttl

        if explicit_max_ttl:
            params['explicit_max_ttl'] = explicit_max_ttl

        if period:
            params['period'] = period

        if orphan:
            return await (await self._post('/v1/auth/token/create-orphan', json=params, wrap_ttl=wrap_ttl)).json()
        elif role:
            return await (await self._post('/v1/auth/token/create/{0}'.format(role), json=params, wrap_ttl=wrap_ttl)).json()
        else:
            return await (await self._post('/v1/auth/token/create', json=params, wrap_ttl=wrap_ttl)).json()

    async def lookup_token(self, token=None, accessor=False, wrap_ttl=None):
        """
        GET /auth/token/lookup/<token>
        GET /auth/token/lookup-accessor/<token-accessor>
        GET /auth/token/lookup-self
        """
        token_param = {
            'token': token,
        }
        accessor_param = {
            'accessor': token,
        }
        if token:
            if accessor:
                path = '/v1/auth/token/lookup-accessor'
                return await (await self._post(path, json=accessor_param, wrap_ttl=wrap_ttl)).json()
            else:
                path = '/v1/auth/token/lookup'
                return await (await self._post(path, json=token_param)).json()
        else:
            path = '/v1/auth/token/lookup-self'
            return await (await self._get(path, wrap_ttl=wrap_ttl)).json()

    def revoke_token(self, token, orphan=False, accessor=False):
        """
        POST /auth/token/revoke
        POST /auth/token/revoke-orphan
        POST /auth/token/revoke-accessor
        """
        if accessor and orphan:
            msg = "revoke_token does not support 'orphan' and 'accessor' flags together"
            raise exceptions.InvalidRequest(msg)
        elif accessor:
            params = {'accessor': token}
            return self._post('/v1/auth/token/revoke-accessor', json=params)
        elif orphan:
            params = {'token': token}
            return self._post('/v1/auth/token/revoke-orphan', json=params)
        else:
            params = {'token': token}
            return self._post('/v1/auth/token/revoke', json=params)

    async def revoke_token_prefix(self, prefix):
        """
        POST /auth/token/revoke-prefix/<prefix>
        """
        return self._post('/v1/auth/token/revoke-prefix/{0}'.format(prefix))

    async def renew_token(self, token=None, increment=None, wrap_ttl=None):
        """
        POST /auth/token/renew/<token>
        POST /auth/token/renew-self
        """
        params = {
            'increment': increment,
        }

        if token:
            path = '/v1/auth/token/renew/{0}'.format(token)
            return await (await self._post(path, json=params, wrap_ttl=wrap_ttl)).json()
        else:
            return await (await self._post('/v1/auth/token/renew-self', json=params, wrap_ttl=wrap_ttl)).json()

    def create_token_role(self, role,
                          allowed_policies=None, disallowed_policies=None,
                          orphan=None, period=None, renewable=None,
                          path_suffix=None, explicit_max_ttl=None):
        """
        POST /auth/token/roles/<role>
        """
        params = {
            'allowed_policies': allowed_policies,
            'disallowed_policies': disallowed_policies,
            'orphan': orphan,
            'period': period,
            'renewable': renewable,
            'path_suffix': path_suffix,
            'explicit_max_ttl': explicit_max_ttl
        }
        return self._post('/v1/auth/token/roles/{0}'.format(role), json=params)

    def token_role(self, role):
        """
        Returns the named token role.
        """
        return self.read('auth/token/roles/{0}'.format(role))

    def delete_token_role(self, role):
        """
        Deletes the named token role.
        """
        return self.delete('auth/token/roles/{0}'.format(role))

    def list_token_roles(self):
        """
        GET /auth/token/roles?list=true
        """
        return self.list('auth/token/roles')

    def logout(self, revoke_token=False):
        """
        Clears the token used for authentication, optionally revoking it before doing so
        """
        if revoke_token:
            return self.revoke_self_token()

        self.token = None

    async def is_authenticated(self):
        """
        Helper method which returns the authentication status of the client
        """
        if not self.token:
            return False

        try:
            await self.lookup_token()
            return True
        except exceptions.Forbidden:
            return False
        except exceptions.InvalidPath:
            return False
        except exceptions.InvalidRequest:
            return False

    def auth_app_id(self, app_id, user_id, mount_point='app-id', use_token=True):
        """
        POST /auth/<mount point>/login
        """
        params = {
            'app_id': app_id,
            'user_id': user_id,
        }

        return self.auth('/v1/auth/{0}/login'.format(mount_point), json=params, use_token=use_token)

    def auth_tls(self, mount_point='cert', use_token=True):
        """
        POST /auth/<mount point>/login
        """
        return self.auth('/v1/auth/{0}/login'.format(mount_point), use_token=use_token)

    def auth_userpass(self, username, password, mount_point='userpass', use_token=True, **kwargs):
        """
        POST /auth/<mount point>/login/<username>
        """
        params = {
            'password': password,
        }

        params.update(kwargs)

        return self.auth('/v1/auth/{0}/login/{1}'.format(mount_point, username), json=params, use_token=use_token)

    async def auth_aws_iam(self, access_key, secret_key, session_token=None, header_value=None, mount_point='aws', role='', use_token=True):
        """
        POST /auth/<mount point>/login
        """
        method = 'POST'
        url = 'https://sts.amazonaws.com/'
        headers = {'Content-Type': 'application/x-www-form-urlencoded; charset=utf-8', 'Host': 'sts.amazonaws.com'}
        body = 'Action=GetCallerIdentity&Version=2011-06-15'

        if header_value:
            headers['X-Vault-AWS-IAM-Server-ID'] = header_value

        auth = aws_utils.SigV4Auth(access_key, secret_key, session_token)
        auth.add_auth(method, headers, body)

        # https://github.com/hashicorp/vault/blob/master/builtin/credential/aws/cli.go
        headers = json.dumps({k: [headers[k]] for k in headers})
        params = {
            'iam_http_request_method': method,
            'iam_request_url': b64encode(url.encode('utf-8')).decode('utf-8'),
            'iam_request_headers': b64encode(headers.encode('utf-8')).decode('utf-8'),
            'iam_request_body': b64encode(body.encode('utf-8')).decode('utf-8'),
            'role': role,
        }

        return await self.auth('/v1/auth/{0}/login'.format(mount_point), json=params, use_token=use_token)

    def auth_ec2(self, pkcs7, nonce=None, role=None, use_token=True, mount_point='aws-ec2'):
        """
        POST /auth/<mount point>/login
        :param pkcs7: str, PKCS#7 version of an AWS Instance Identity Document from the EC2 Metadata Service.
        :param nonce: str, optional nonce returned as part of the original authentication request. Not required if the
         backend has "allow_instance_migration" or "disallow_reauthentication" options turned on.
        :param role: str, identifier for the AWS auth backend role being requested
        :param use_token: bool, if True, uses the token in the response received from the auth request to set the "token"
         attribute on the current Client class instance.
        :param mount_point: str, The "path" the AWS auth backend was mounted on. Vault currently defaults to "aws".
         "aws-ec2" is the default argument for backwards comparability within this module.
        :return: dict, parsed JSON response from the auth POST request
        """
        params = {'pkcs7': pkcs7}
        if nonce:
            params['nonce'] = nonce
        if role:
            params['role'] = role

        return self.auth('/v1/auth/{0}/login'.format(mount_point), json=params, use_token=use_token)

    def auth_gcp(self, role, jwt, mount_point='gcp', use_token=True):
        """
        POST /auth/<mount point>/login
        :param role: str, identifier for the GCP auth backend role being requested
        :param jwt: str, JSON Web Token from the GCP metadata service
        :param mount_point: str, The "path" the GCP auth backend was mounted on. Vault currently defaults to "gcp".
        :param use_token: bool, if True, uses the token in the response received from the auth request to set the "token"
        attribute on the current Client class instance.
        :return: dict, parsed JSON response from the auth POST request
        """

        params = {
            'role': role,
            'jwt': jwt
        }

        return self.auth('/v1/auth/{0}/login'.format(mount_point), json=params, use_token=use_token)

    def create_userpass(self, username, password, policies, mount_point='userpass', **kwargs):
        """
        POST /auth/<mount point>/users/<username>
        """

        # Users can have more than 1 policy. It is easier for the user to pass in the
        # policies as a list so if they do, we need to convert to a , delimited string.
        if isinstance(policies, (list, set, tuple)):
            policies = ','.join(policies)

        params = {
            'password': password,
            'policies': policies
        }
        params.update(kwargs)

        return self._post('/v1/auth/{}/users/{}'.format(mount_point, username), json=params)

    async def list_userpass(self, mount_point='userpass'):
        """
        GET /auth/<mount point>/users?list=true
        """
        try:
            return await (await self._get('/v1/auth/{}/users'.format(mount_point), params={'list': 'true'})).json()
        except exceptions.InvalidPath:
            return None

    async def read_userpass(self, username, mount_point='userpass'):
        """
        GET /auth/<mount point>/users/<username>
        """
        return await (await self._get('/v1/auth/{}/users/{}'.format(mount_point, username))).json()

    def update_userpass_policies(self, username, policies, mount_point='userpass'):
        """
        POST /auth/<mount point>/users/<username>/policies
        """
        # userpass can have more than 1 policy. It is easier for the user to pass in the
        # policies as a list so if they do, we need to convert to a , delimited string.
        if isinstance(policies, (list, set, tuple)):
            policies = ','.join(policies)

        params = {
            'policies': policies
        }

        return self._post('/v1/auth/{}/users/{}/policies'.format(mount_point, username), json=params)

    def update_userpass_password(self, username, password, mount_point='userpass'):
        """
        POST /auth/<mount point>/users/<username>/password
        """
        params = {
            'password': password
        }
        return self._post('/v1/auth/{}/users/{}/password'.format(mount_point, username), json=params)

    def delete_userpass(self, username, mount_point='userpass'):
        """
        DELETE /auth/<mount point>/users/<username>
        """
        return self._delete('/v1/auth/{}/users/{}'.format(mount_point, username))

    def create_app_id(self, app_id, policies, display_name=None, mount_point='app-id', **kwargs):
        """
        POST /auth/<mount point>/map/app-id/<app_id>
        """

        # app-id can have more than 1 policy. It is easier for the user to pass in the
        # policies as a list so if they do, we need to convert to a , delimited string.
        if isinstance(policies, (list, set, tuple)):
            policies = ','.join(policies)

        params = {
            'value': policies
        }

        # Only use the display_name if it has a value. Made it a named param for user
        # convienence instead of leaving it as part of the kwargs
        if display_name:
            params['display_name'] = display_name

        params.update(kwargs)

        return self._post('/v1/auth/{}/map/app-id/{}'.format(mount_point, app_id), json=params)

    async def get_app_id(self, app_id, mount_point='app-id', wrap_ttl=None):
        """
        GET /auth/<mount_point>/map/app-id/<app_id>
        """
        path = '/v1/auth/{0}/map/app-id/{1}'.format(mount_point, app_id)
        return await (await self._get(path, wrap_ttl=wrap_ttl)).json()

    def delete_app_id(self, app_id, mount_point='app-id'):
        """
        DELETE /auth/<mount_point>/map/app-id/<app_id>
        """
        return self._delete('/v1/auth/{0}/map/app-id/{1}'.format(mount_point, app_id))

    def create_user_id(self, user_id, app_id, cidr_block=None, mount_point='app-id', **kwargs):
        """
        POST /auth/<mount point>/map/user-id/<user_id>
        """

        # user-id can be associated to more than 1 app-id (aka policy). It is easier for the user to
        # pass in the policies as a list so if they do, we need to convert to a , delimited string.
        if isinstance(app_id, (list, set, tuple)):
            app_id = ','.join(app_id)

        params = {
            'value': app_id
        }

        # Only use the cidr_block if it has a value. Made it a named param for user
        # convienence instead of leaving it as part of the kwargs
        if cidr_block:
            params['cidr_block'] = cidr_block

        params.update(kwargs)

        return self._post('/v1/auth/{}/map/user-id/{}'.format(mount_point, user_id), json=params)

    async def get_user_id(self, user_id, mount_point='app-id', wrap_ttl=None):
        """
        GET /auth/<mount_point>/map/user-id/<user_id>
        """
        path = '/v1/auth/{0}/map/user-id/{1}'.format(mount_point, user_id)
        return await (await self._get(path, wrap_ttl=wrap_ttl)).json()

    def delete_user_id(self, user_id, mount_point='app-id'):
        """
        DELETE /auth/<mount_point>/map/user-id/<user_id>
        """
        return self._delete('/v1/auth/{0}/map/user-id/{1}'.format(mount_point, user_id))

    def create_vault_ec2_client_configuration(self, access_key, secret_key, endpoint=None, mount_point='aws-ec2'):
        """
        POST /auth/<mount_point>/config/client
        """
        params = {
            'access_key': access_key,
            'secret_key': secret_key
        }
        if endpoint is not None:
            params['endpoint'] = endpoint

        return self._post('/v1/auth/{0}/config/client'.format(mount_point), json=params)

    async def get_vault_ec2_client_configuration(self, mount_point='aws-ec2'):
        """
        GET /auth/<mount_point>/config/client
        """
        return await (await self._get('/v1/auth/{0}/config/client'.format(mount_point))).json()

    def delete_vault_ec2_client_configuration(self, mount_point='aws-ec2'):
        """
        DELETE /auth/<mount_point>/config/client
        """
        return self._delete('/v1/auth/{0}/config/client'.format(mount_point))

    def create_vault_ec2_certificate_configuration(self, cert_name, aws_public_cert, mount_point='aws-ec2'):
        """
        POST /auth/<mount_point>/config/certificate/<cert_name>
        """
        params = {
            'cert_name': cert_name,
            'aws_public_cert': aws_public_cert
        }
        return self._post('/v1/auth/{0}/config/certificate/{1}'.format(mount_point, cert_name), json=params)

    async def get_vault_ec2_certificate_configuration(self, cert_name, mount_point='aws-ec2'):
        """
        GET /auth/<mount_point>/config/certificate/<cert_name>
        """
        return await (await self._get('/v1/auth/{0}/config/certificate/{1}'.format(mount_point, cert_name))).json()

    async def list_vault_ec2_certificate_configurations(self, mount_point='aws-ec2'):
        """
        GET /auth/<mount_point>/config/certificates?list=true
        """
        params = {'list': 'true'}
        return await (await self._get('/v1/auth/{0}/config/certificates'.format(mount_point), params=params)).json()

    def create_ec2_role(self, role, bound_ami_id=None, bound_account_id=None, bound_iam_role_arn=None,
                        bound_iam_instance_profile_arn=None, bound_ec2_instance_id=None, bound_region=None,
                        bound_vpc_id=None, bound_subnet_id=None, role_tag=None,  ttl=None, max_ttl=None, period=None,
                        policies=None, allow_instance_migration=False, disallow_reauthentication=False,
                        resolve_aws_unique_ids=None, mount_point='aws-ec2'):
        """
        POST /auth/<mount_point>/role/<role>
        """
        params = {
            'role': role,
            'auth_type': 'ec2',
            'disallow_reauthentication': disallow_reauthentication,
            'allow_instance_migration': allow_instance_migration
        }

        if bound_ami_id is not None:
            params['bound_ami_id'] = bound_ami_id
        if bound_account_id is not None:
            params['bound_account_id'] = bound_account_id
        if bound_iam_role_arn is not None:
            params['bound_iam_role_arn'] = bound_iam_role_arn
        if bound_ec2_instance_id is not None:
            params['bound_iam_instance_profile_arn'] = bound_ec2_instance_id
        if bound_iam_instance_profile_arn is not None:
            params['bound_iam_instance_profile_arn'] = bound_iam_instance_profile_arn
        if bound_region is not None:
            params['bound_region'] = bound_region
        if bound_vpc_id is not None:
            params['bound_vpc_id'] = bound_vpc_id
        if bound_subnet_id is not None:
            params['bound_subnet_id'] = bound_subnet_id
        if role_tag is not None:
            params['role_tag'] = role_tag
        if ttl is not None:
            params['ttl'] = ttl
        else:
            params['ttl'] = 0
        if max_ttl is not None:
            params['max_ttl'] = max_ttl
        else:
            params['max_ttl'] = 0
        if period is not None:
            params['period'] = period
        else:
            params['period'] = 0
        if policies is not None:
            params['policies'] = policies
        if resolve_aws_unique_ids is not None:
            params['resolve_aws_unique_ids'] = resolve_aws_unique_ids

        return self._post('/v1/auth/{0}/role/{1}'.format(mount_point, role), json=params)

    async def get_ec2_role(self, role, mount_point='aws-ec2'):
        """
        GET /auth/<mount_point>/role/<role>
        """
        return await (await self._get('/v1/auth/{0}/role/{1}'.format(mount_point, role))).json()

    def delete_ec2_role(self, role, mount_point='aws-ec2'):
        """
        DELETE /auth/<mount_point>/role/<role>
        """
        return self._delete('/v1/auth/{0}/role/{1}'.format(mount_point, role))

    async def list_ec2_roles(self, mount_point='aws-ec2'):
        """
        GET /auth/<mount_point>/roles?list=true
        """
        try:
            return await (await self._get('/v1/auth/{0}/roles'.format(mount_point), params={'list': 'true'})).json()
        except exceptions.InvalidPath:
            return None

    def create_ec2_role_tag(self, role, policies=None, max_ttl=None, instance_id=None,
                            disallow_reauthentication=False, allow_instance_migration=False, mount_point='aws-ec2'):
        """
        POST /auth/<mount_point>/role/<role>/tag
        """
        params = {
            'role': role,
            'disallow_reauthentication': disallow_reauthentication,
            'allow_instance_migration': allow_instance_migration
        }

        if max_ttl is not None:
            params['max_ttl'] = max_ttl
        if policies is not None:
            params['policies'] = policies
        if instance_id is not None:
            params['instance_id'] = instance_id
        return self._post('/v1/auth/{0}/role/{1}/tag'.format(mount_point, role), json=params)

    def auth_ldap(self, username, password, mount_point='ldap', use_token=True, **kwargs):
        """
        POST /auth/<mount point>/login/<username>
        """
        params = {
            'password': password,
        }

        params.update(kwargs)

        return self.auth('/v1/auth/{0}/login/{1}'.format(mount_point, username), json=params, use_token=use_token)

    def auth_github(self, token, mount_point='github', use_token=True):
        """
        POST /auth/<mount point>/login
        """
        params = {
            'token': token,
        }

        return self.auth('/v1/auth/{0}/login'.format(mount_point), json=params, use_token=use_token)

    def auth_cubbyhole(self, token):
        """
        POST /v1/sys/wrapping/unwrap
        """
        self.token = token
        return self.auth('/v1/sys/wrapping/unwrap')

    async def auth(self, url, use_token=True, **kwargs):
        response = await (await self._post(url, **kwargs)).json()

        if use_token:
            self.token = response['auth']['client_token']

        return response

    async def list_auth_backends(self):
        """
        GET /sys/auth
        """
        return await (await self._get('/v1/sys/auth')).json()

    def enable_auth_backend(self, backend_type, description=None, mount_point=None):
        """
        POST /sys/auth/<mount point>
        """
        if not mount_point:
            mount_point = backend_type

        params = {
            'type': backend_type,
            'description': description,
        }

        return self._post('/v1/sys/auth/{0}'.format(mount_point), json=params)

    def tune_auth_backend(self, backend_type, mount_point=None, default_lease_ttl=None, max_lease_ttl=None, description=None,
                          audit_non_hmac_request_keys=None, audit_non_hmac_response_keys=None, listing_visibility=None,
                          passthrough_request_headers=None):
        """
        POST /sys/auth/<mount point>/tune
        :param backend_type: str, Name of the auth backend to modify (e.g., token, approle, etc.)
        :param mount_point: str, The path the associated auth backend is mounted under.
        :param description: str, Specifies the description of the mount. This overrides the current stored value, if any.
        :param default_lease_ttl: int,
        :param max_lease_ttl: int,
        :param audit_non_hmac_request_keys: list, Specifies the comma-separated list of keys that will not be HMAC'd by
        audit devices in the request data object.
        :param audit_non_hmac_response_keys: list, Specifies the comma-separated list of keys that will not be HMAC'd
        by audit devices in the response data object.
        :param listing_visibility: str, Speficies whether to show this mount in the UI-specific listing endpoint.
        Valid values are "unauth" or "".
        :param passthrough_request_headers: list, Comma-separated list of headers to whitelist and pass from the request
        to the backend.
        :return: dict, The JSON response from Vault
        """
        if not mount_point:
            mount_point = backend_type
        # All parameters are optional for this method. Until/unless we include input validation, we simply loop over the
        # parameters and add which parameters are set.
        optional_parameters = [
            'default_lease_ttl',
            'max_lease_ttl',
            'description',
            'audit_non_hmac_request_keys',
            'audit_non_hmac_response_keys',
            'listing_visibility',
            'passthrough_request_headers',
        ]
        params = {}
        for optional_parameter in optional_parameters:
            if locals().get(optional_parameter) is not None:
                params[optional_parameter] = locals().get(optional_parameter)
        return self._post('/v1/sys/auth/{0}/tune'.format(mount_point), json=params)

    async def get_auth_backend_tuning(self, backend_type, mount_point=None):
        """
        GET /sys/auth/<mount point>/tune
        :param backend_type: str, Name of the auth backend to modify (e.g., token, approle, etc.)
        :param mount_point: str, The path the associated auth backend is mounted under.
        :return: dict, The JSON response from Vault
        """
        if not mount_point:
            mount_point = backend_type

        return await (await self._get('/v1/sys/auth/{0}/tune'.format(mount_point))).json()

    def disable_auth_backend(self, mount_point):
        """
        DELETE /sys/auth/<mount point>
        """
        return self._delete('/v1/sys/auth/{0}'.format(mount_point))

    def create_role(self, role_name, mount_point='approle', **kwargs):
        """
        POST /auth/<mount_point>/role/<role name>
        """

        return self._post('/v1/auth/{0}/role/{1}'.format(mount_point, role_name), json=kwargs)

    def delete_role(self, role_name, mount_point='approle'):
        """
        DELETE /auth/<mount_point>/role/<role name>
        """

        return self._delete('/v1/auth/{0}/role/{1}'.format(mount_point, role_name))

    async def list_roles(self, mount_point='approle'):
        """
        GET /auth/<mount_point>/role
        """

        return await (await self._get('/v1/auth/{0}/role?list=true'.format(mount_point))).json()

    async def get_role_id(self, role_name, mount_point='approle'):
        """
        GET /auth/<mount_point>/role/<role name>/role-id
        """

        url = '/v1/auth/{0}/role/{1}/role-id'.format(mount_point, role_name)
        return (await (await self._get(url)).json())['data']['role_id']

    def set_role_id(self, role_name, role_id, mount_point='approle'):
        """
        POST /auth/<mount_point>/role/<role name>/role-id
        """

        url = '/v1/auth/{0}/role/{1}/role-id'.format(mount_point, role_name)
        params = {
            'role_id': role_id
        }
        return self._post(url, json=params)

    async def get_role(self, role_name, mount_point='approle'):
        """
        GET /auth/<mount_point>/role/<role name>
        """
        return await (await self._get('/v1/auth/{0}/role/{1}'.format(mount_point, role_name))).json()

    async def create_role_secret_id(self, role_name, meta=None, cidr_list=None, wrap_ttl=None, mount_point='approle'):
        """
        POST /auth/<mount_point>/role/<role name>/secret-id
        """

        url = '/v1/auth/{0}/role/{1}/secret-id'.format(mount_point, role_name)
        params = {}
        if meta is not None:
            params['metadata'] = json.dumps(meta)
        if cidr_list is not None:
            params['cidr_list'] = cidr_list
        return await (await self._post(url, json=params, wrap_ttl=wrap_ttl)).json()

    async def get_role_secret_id(self, role_name, secret_id, mount_point='approle'):
        """
        POST /auth/<mount_point>/role/<role name>/secret-id/lookup
        """
        url = '/v1/auth/{0}/role/{1}/secret-id/lookup'.format(mount_point, role_name)
        params = {
            'secret_id': secret_id
        }
        return await (await self._post(url, json=params)).json()

    async def list_role_secrets(self, role_name, mount_point='approle'):
        """
        GET /auth/<mount_point>/role/<role name>/secret-id?list=true
        """
        url = '/v1/auth/{0}/role/{1}/secret-id?list=true'.format(mount_point, role_name)
        return await (await self._get(url)).json()

    async def get_role_secret_id_accessor(self, role_name, secret_id_accessor, mount_point='approle'):
        """
        POST /auth/<mount_point>/role/<role name>/secret-id-accessor/lookup
        """
        url = '/v1/auth/{0}/role/{1}/secret-id-accessor/lookup'.format(mount_point, role_name)
        params = {'secret_id_accessor': secret_id_accessor}
        return await (await self._post(url, json=params)).json()

    def delete_role_secret_id(self, role_name, secret_id, mount_point='approle'):
        """
        POST /auth/<mount_point>/role/<role name>/secret-id/destroy
        """
        url = '/v1/auth/{0}/role/{1}/secret-id/destroy'.format(mount_point, role_name)
        params = {
            'secret_id': secret_id
        }
        return self._post(url, json=params)

    def delete_role_secret_id_accessor(self, role_name, secret_id_accessor, mount_point='approle'):
        """
        DELETE /auth/<mount_point>/role/<role name>/secret-id/<secret_id_accessor>
        """
        url = '/v1/auth/{0}/role/{1}/secret-id-accessor/{2}'.format(mount_point, role_name, secret_id_accessor)
        return self._delete(url)

    async def create_role_custom_secret_id(self, role_name, secret_id, meta=None, mount_point='approle'):
        """
        POST /auth/<mount_point>/role/<role name>/custom-secret-id
        """
        url = '/v1/auth/{0}/role/{1}/custom-secret-id'.format(mount_point, role_name)
        params = {
            'secret_id': secret_id
        }
        if meta is not None:
            params['meta'] = meta
        return await (await self._post(url, json=params)).json()

    def auth_approle(self, role_id, secret_id=None, mount_point='approle', use_token=True):
        """
        POST /auth/<mount_point>/login
        """
        params = {
            'role_id': role_id
        }
        if secret_id is not None:
            params['secret_id'] = secret_id

        return self.auth('/v1/auth/{0}/login'.format(mount_point), json=params, use_token=use_token)

    def create_kubernetes_configuration(self, kubernetes_host, kubernetes_ca_cert=None, token_reviewer_jwt=None, pem_keys=None, mount_point='kubernetes'):
        """
        POST /auth/<mount_point>/config
        :param kubernetes_host: str, a host:port pair, or a URL to the base of the Kubernetes API server.
        :param kubernetes_ca_cert: str, PEM encoded CA cert for use by the TLS client used to talk with the Kubernetes API.
        :param token_reviewer_jwt: str, A service account JWT used to access the TokenReview API to validate other
        JWTs during login. If not set the JWT used for login will be used to access the API.
        :param pem_keys: list, Optional list of PEM-formated public keys or certificates used to verify the signatures of
        Kubernetes service account JWTs. If a certificate is given, its public key will be extracted. Not every
        installation of Kubernetes exposes these keys.
        :param mount_point: str, The "path" the k8s auth backend was mounted on. Vault currently defaults to "kubernetes".
        :return: requests.Response, will be an empty body with a 204 status code upon success
        """
        params = {
            'kubernetes_host': kubernetes_host,
            'kubernetes_ca_cert': kubernetes_ca_cert,
        }

        if token_reviewer_jwt is not None:
            params['token_reviewer_jwt'] = token_reviewer_jwt
        if pem_keys is not None:
            params['pem_keys'] = pem_keys

        url = 'v1/auth/{0}/config'.format(mount_point)
        return self._post(url, json=params)

    async def get_kubernetes_configuration(self, mount_point='kubernetes'):
        """
        GET /auth/<mount_point>/config
        :param mount_point: str, The "path" the k8s auth backend was mounted on. Vault currently defaults to "kubernetes".
        :return: dict, parsed JSON response from the config GET request
        """

        url = '/v1/auth/{0}/config'.format(mount_point)
        return await (await self._get(url)).json()

    def create_kubernetes_role(self, name, bound_service_account_names, bound_service_account_namespaces, ttl="",
                               max_ttl="", period="", policies=None, mount_point='kubernetes'):
        """
        POST /auth/<mount_point>/role/:name
        :param name: str, Name of the role.
        :param bound_service_account_names: list, List of service account names able to access this role. If set to "*" all
        names are allowed, both this and bound_service_account_namespaces can not be "*".
        :param bound_service_account_namespaces: list, List of namespaces allowed to access this role. If set to "*" all
        namespaces are allowed, both this and bound_service_account_names can not be set to "*".
        :param ttl: str, The TTL period of tokens issued using this role in seconds.
        :param max_ttl: str, The maximum allowed lifetime of tokens issued in seconds using this role.
        :param period: str, If set, indicates that the token generated using this role should never expire.
        The token should be renewed within the duration specified by this value. At each renewal, the token's TTL will
        be set to the value of this parameter.
        :param policies: list, Policies to be set on tokens issued using this role
        :param mount_point: str, The "path" the k8s auth backend was mounted on. Vault currently defaults to "kubernetes".
        :return: requests.Response, will be an empty body with a 204 status code upon success
        """
        if bound_service_account_names == '*' and bound_service_account_namespaces == '*':
            error_message = 'bound_service_account_names and bound_service_account_namespaces can not both be set to "*"'
            raise exceptions.ParamValidationError(error_message)

        params = {
            'bound_service_account_names': bound_service_account_names,
            'bound_service_account_namespaces': bound_service_account_namespaces,
            'ttl': ttl,
            'max_ttl': max_ttl,
            'period': period,
            'policies': policies,
        }
        url = 'v1/auth/{0}/role/{1}'.format(mount_point, name)
        return self._post(url, json=params)

    async def get_kubernetes_role(self, name, mount_point='kubernetes'):
        """
        GET /auth/<mount_point>/role/:name
        :param name: str, Name of the role.
        :param mount_point: str, The "path" the k8s auth backend was mounted on. Vault currently defaults to "kubernetes".
        :return: dict, parsed JSON response from the read role GET request
        """

        url = 'v1/auth/{0}/role/{1}'.format(mount_point, name)
        return await (await self._get(url)).json()

    async def list_kubernetes_roles(self, mount_point='kubernetes'):
        """
        GET /auth/<mount_point>/role?list=true
        :param mount_point: str, The "path" the k8s auth backend was mounted on. Vault currently defaults to "kubernetes".
        :return: dict, parsed JSON response from the list roles GET request
        """

        url = 'v1/auth/{0}/role?list=true'.format(mount_point)
        return await (await self._get(url)).json()

    def delete_kubernetes_role(self, role, mount_point='kubernetes'):
        """
        DELETE /auth/<mount_point>/role/:role
        :param role: str, Name of the role.
        :param mount_point: str, The "path" the k8s auth backend was mounted on. Vault currently defaults to "kubernetes".
        :return: requests.Response, will be an empty body with a 204 status code upon success
        """

        url = 'v1/auth/{0}/role/{1}'.format(mount_point, role)
        return self._delete(url)

    def auth_kubernetes(self, role, jwt, use_token=True, mount_point='kubernetes'):
        """
        POST /auth/<mount_point>/login
        :param role: str, Name of the role against which the login is being attempted.
        :param jwt: str, Signed JSON Web Token (JWT) for authenticating a service account.
        :param use_token: bool, if True, uses the token in the response received from the auth request to set the "token"
         attribute on the current Client class instance.
        :param mount_point: str, The "path" the k8s auth backend was mounted on. Vault currently defaults to "kubernetes".
        :return: dict, parsed JSON response from the config POST request
        """
        params = {
            'role': role,
            'jwt': jwt
        }
        url = 'v1/auth/{0}/login'.format(mount_point)
        return self.auth(url, json=params, use_token=use_token)

    def transit_create_key(self, name, convergent_encryption=None, derived=None, exportable=None,
                           key_type=None, mount_point='transit'):
        """
        POST /<mount_point>/keys/<name>
        """
        url = '/v1/{0}/keys/{1}'.format(mount_point, name)
        params = {}
        if convergent_encryption is not None:
            params['convergent_encryption'] = convergent_encryption
        if derived is not None:
            params['derived'] = derived
        if exportable is not None:
            params['exportable'] = exportable
        if key_type is not None:
            params['type'] = key_type

        return self._post(url, json=params)

    async def transit_read_key(self, name, mount_point='transit'):
        """
        GET /<mount_point>/keys/<name>
        """
        url = '/v1/{0}/keys/{1}'.format(mount_point, name)
        return await (await self._get(url)).json()

    async def transit_list_keys(self, mount_point='transit'):
        """
        GET /<mount_point>/keys?list=true
        """
        url = '/v1/{0}/keys?list=true'.format(mount_point)
        return await (await self._get(url)).json()

    def transit_delete_key(self, name, mount_point='transit'):
        """
        DELETE /<mount_point>/keys/<name>
        """
        url = '/v1/{0}/keys/{1}'.format(mount_point, name)
        return self._delete(url)

    def transit_update_key(self, name, min_decryption_version=None, min_encryption_version=None, deletion_allowed=None,
                           mount_point='transit'):
        """
        POST /<mount_point>/keys/<name>/config
        """
        url = '/v1/{0}/keys/{1}/config'.format(mount_point, name)
        params = {}
        if min_decryption_version is not None:
            params['min_decryption_version'] = min_decryption_version
        if min_encryption_version is not None:
            params['min_encryption_version'] = min_encryption_version
        if deletion_allowed is not None:
            params['deletion_allowed'] = deletion_allowed

        return self._post(url, json=params)

    def transit_rotate_key(self, name, mount_point='transit'):
        """
        POST /<mount_point>/keys/<name>/rotate
        """
        url = '/v1/{0}/keys/{1}/rotate'.format(mount_point, name)
        return self._post(url)

    async def transit_export_key(self, name, key_type, version=None, mount_point='transit'):
        """
        GET /<mount_point>/export/<key_type>/<name>(/<version>)
        """
        if version is not None:
            url = '/v1/{0}/export/{1}/{2}/{3}'.format(mount_point, key_type, name, version)
        else:
            url = '/v1/{0}/export/{1}/{2}'.format(mount_point, key_type, name)
        return await (await self._get(url)).json()

    async def transit_encrypt_data(self, name, plaintext, context=None, key_version=None, nonce=None, batch_input=None,
                                   key_type=None, convergent_encryption=None, mount_point='transit'):
        """
        POST /<mount_point>/encrypt/<name>
        """
        url = '/v1/{0}/encrypt/{1}'.format(mount_point, name)
        params = {
            'plaintext': plaintext
        }
        if context is not None:
            params['context'] = context
        if key_version is not None:
            params['key_version'] = key_version
        if nonce is not None:
            params['nonce'] = nonce
        if batch_input is not None:
            params['batch_input'] = batch_input
        if key_type is not None:
            params['type'] = key_type
        if convergent_encryption is not None:
            params['convergent_encryption'] = convergent_encryption

        return await (await self._post(url, json=params)).json()

    async def transit_decrypt_data(self, name, ciphertext, context=None, nonce=None, batch_input=None, mount_point='transit'):
        """
        POST /<mount_point>/decrypt/<name>
        """
        url = '/v1/{0}/decrypt/{1}'.format(mount_point, name)
        params = {
            'ciphertext': ciphertext
        }
        if context is not None:
            params['context'] = context
        if nonce is not None:
            params['nonce'] = nonce
        if batch_input is not None:
            params['batch_input'] = batch_input

        return await (await self._post(url, json=params)).json()

    async def transit_rewrap_data(self, name, ciphertext, context=None, key_version=None, nonce=None, batch_input=None,
                                  mount_point='transit'):
        """
        POST /<mount_point>/rewrap/<name>
        """
        url = '/v1/{0}/rewrap/{1}'.format(mount_point, name)
        params = {
            'ciphertext': ciphertext
        }
        if context is not None:
            params['context'] = context
        if key_version is not None:
            params['key_version'] = key_version
        if nonce is not None:
            params['nonce'] = nonce
        if batch_input is not None:
            params['batch_input'] = batch_input

        return await (await self._post(url, json=params)).json()

    async def transit_generate_data_key(self, name, key_type, context=None, nonce=None, bits=None, mount_point='transit'):
        """
        POST /<mount_point>/datakey/<type>/<name>
        """
        url = '/v1/{0}/datakey/{1}/{2}'.format(mount_point, key_type, name)
        params = {}
        if context is not None:
            params['context'] = context
        if nonce is not None:
            params['nonce'] = nonce
        if bits is not None:
            params['bits'] = bits

        return await (await self._post(url, json=params)).json()

    async def transit_generate_rand_bytes(self, data_bytes=None, output_format=None, mount_point='transit'):
        """
        POST /<mount_point>/random(/<data_bytes>)
        """
        if data_bytes is not None:
            url = '/v1/{0}/random/{1}'.format(mount_point, data_bytes)
        else:
            url = '/v1/{0}/random'.format(mount_point)

        params = {}
        if output_format is not None:
            params["format"] = output_format

        return await (await self._post(url, json=params)).json()

    async def transit_hash_data(self, hash_input, algorithm=None, output_format=None, mount_point='transit'):
        """
        POST /<mount_point>/hash(/<algorithm>)
        """
        if algorithm is not None:
            url = '/v1/{0}/hash/{1}'.format(mount_point, algorithm)
        else:
            url = '/v1/{0}/hash'.format(mount_point)

        params = {
            'input': hash_input
        }
        if output_format is not None:
            params['format'] = output_format

        return await (await self._post(url, json=params)).json()

    async def transit_generate_hmac(self, name, hmac_input, key_version=None, algorithm=None, mount_point='transit'):
        """
        POST /<mount_point>/hmac/<name>(/<algorithm>)
        """
        if algorithm is not None:
            url = '/v1/{0}/hmac/{1}/{2}'.format(mount_point, name, algorithm)
        else:
            url = '/v1/{0}/hmac/{1}'.format(mount_point, name)
        params = {
            'input': hmac_input
        }
        if key_version is not None:
            params['key_version'] = key_version

        return await (await self._post(url, json=params)).json()

    async def transit_sign_data(self, name, input_data, key_version=None, algorithm=None, context=None, prehashed=None,
                                mount_point='transit', signature_algorithm='pss'):
        """
        POST /<mount_point>/sign/<name>(/<algorithm>)
        """
        if algorithm is not None:
            url = '/v1/{0}/sign/{1}/{2}'.format(mount_point, name, algorithm)
        else:
            url = '/v1/{0}/sign/{1}'.format(mount_point, name)

        params = {
            'input': input_data
        }
        if key_version is not None:
            params['key_version'] = key_version
        if context is not None:
            params['context'] = context
        if prehashed is not None:
            params['prehashed'] = prehashed
        params['signature_algorithm'] = signature_algorithm

        return await (await self._post(url, json=params)).json()

    async def transit_verify_signed_data(self, name, input_data, algorithm=None, signature=None, hmac=None, context=None,
                                         prehashed=None, mount_point='transit', signature_algorithm='pss'):
        """
        POST /<mount_point>/verify/<name>(/<algorithm>)
        """
        if algorithm is not None:
            url = '/v1/{0}/verify/{1}/{2}'.format(mount_point, name, algorithm)
        else:
            url = '/v1/{0}/verify/{1}'.format(mount_point, name)

        params = {
            'input': input_data
        }
        if signature is not None:
            params['signature'] = signature
        if hmac is not None:
            params['hmac'] = hmac
        if context is not None:
            params['context'] = context
        if prehashed is not None:
            params['prehashed'] = prehashed
        params['signature_algorithm'] = signature_algorithm

        return await (await self._post(url, json=params)).json()

    def close(self):
        """
        Close the underlying Requests session
        """
        return self.session.close()

    def _get(self, url, **kwargs):
        return self.__request('get', url, **kwargs)

    def _post(self, url, **kwargs):
        return self.__request('post', url, **kwargs)

    def _put(self, url, **kwargs):
        return self.__request('put', url, **kwargs)

    def _delete(self, url, **kwargs):
        return self.__request('delete', url, **kwargs)

    def urljoin(self, *args):
        """
        Joins given arguments into a url. Trailing and leading slashes are
        stripped for each argument.
        """

        return '/'.join(map(lambda x: str(x).strip('/'), args))

    async def __request(self, method, url, headers=None, **kwargs):
        url = self.urljoin(self._url, url)

        if not headers:
            headers = {}

        if self.token:
            headers['X-Vault-Token'] = self.token

        wrap_ttl = kwargs.pop('wrap_ttl', None)
        if wrap_ttl:
            headers['X-Vault-Wrap-TTL'] = str(wrap_ttl)
        response = await self.session.request(
            method, url, headers=headers,
            allow_redirects=False,
            ssl=self._sslcontext,
            proxy=self._proxies, **kwargs)
        while response.status >= 300 and response.status < 400 and self.allow_redirects and 'Location' in response.headers:
            url = self.urljoin(self._url, response.headers['Location'])
            response = await self.session.request(
                method, url, headers=headers,
                allow_redirects=False,
                ssl=self._sslcontext,
                proxy=self._proxies, **kwargs)

        if response.status >= 400 and response.status < 600:
            text = errors = None
            if response.headers.get('Content-Type') == 'application/json':
                errors = (await response.json()).get('errors')
            if errors is None:
                text = response.text
            self.__raise_error(response.status, text, errors=errors)

        return response

    def __raise_error(self, status_code, message=None, errors=None):
        if status_code == 400:
            raise exceptions.InvalidRequest(message, errors=errors)
        elif status_code == 401:
            raise exceptions.Unauthorized(message, errors=errors)
        elif status_code == 403:
            raise exceptions.Forbidden(message, errors=errors)
        elif status_code == 404:
            raise exceptions.InvalidPath(message, errors=errors)
        elif status_code == 429:
            raise exceptions.RateLimitExceeded(message, errors=errors)
        elif status_code == 500:
            raise exceptions.InternalServerError(message, errors=errors)
        elif status_code == 501:
            raise exceptions.VaultNotInitialized(message, errors=errors)
        elif status_code == 503:
            raise exceptions.VaultDown(message, errors=errors)
        else:
            raise exceptions.UnexpectedError(message)



import asyncio
import concurrent

from async_hvac.v1 import AsyncClient


def async_to_sync(self, f):
    def wrapper(*args, **kwargs):
        if self._loop.is_running():
            return f(*args, **kwargs)
        coro = asyncio.coroutine(f)
        future = coro(*args, **kwargs)
        return self._executor.submit(
            self._loop.run_until_complete,
            future).result()
    return wrapper


# Just for test and try and do not work properly in some cases... not part of the Async API
class Client(AsyncClient):

    def __init__(self, url='http://127.0.0.1:8200', token=None,
                 cert=None, verify=True, timeout=30, proxies=None,
                 allow_redirects=True, session=None, sync=True,
                 loop=None):
        super(Client, self).__init__(
            url, token, cert, verify, timeout,
            proxies, allow_redirects, session, loop)
        self._sync = sync
        if sync:
            for attr in AsyncClient.__dict__:
                attr_obj = getattr(AsyncClient, attr)
                if callable(attr_obj) and not attr.startswith('_'):
                    setattr(self, attr, async_to_sync(self, getattr(self, attr)))
            self._executor = concurrent.futures.ThreadPoolExecutor(
                max_workers=3,
            )
            self._loop = asyncio.new_event_loop()
        else:
            if loop:
                self._loop = loop
            else:
                self._loop = asyncio.get_event_loop()

    @property
    def seal_status(self):
        if not self._sync or self._loop.is_running():
            return super(Client, self).seal_status
        return self._executor.submit(
            self._loop.run_until_complete,
            super(Client, self).seal_status).result()

    @property
    def generate_root_status(self):
        if not self._sync or self._loop.is_running():
            return super(Client, self).generate_root_status
        return self._executor.submit(
            self._loop.run_until_complete,
            super(Client, self).generate_root_status).result()

    @property
    def key_status(self):
        if not self._sync or self._loop.is_running():
            return super(Client, self).key_status
        return self._executor.submit(
            self._loop.run_until_complete,
            super(Client, self).key_status).result()

    @property
    def rekey_status(self):
        if not self._sync or self._loop.is_running():
            return super(Client, self).rekey_status
        return self._executor.submit(
            self._loop.run_until_complete,
            super(Client, self).rekey_status).result()

    @property
    def ha_status(self):
        if not self._sync or self._loop.is_running():
            return super(Client, self).ha_status
        return self._executor.submit(
            self._loop.run_until_complete,
            super(Client, self).ha_status).result()



import hmac
from datetime import datetime
from hashlib import sha256


class SigV4Auth(object):
    def __init__(self, access_key, secret_key, session_token=None):
        self.access_key = access_key
        self.secret_key = secret_key
        self.session_token = session_token

    def add_auth(self, method, headers, body):
        if body:
            headers['Content-Length'] = str(len(body))
        timestamp = datetime.utcnow().strftime('%Y%m%dT%H%M%SZ')
        headers['X-Amz-Date'] = timestamp

        if self.session_token:
            headers['X-Amz-Security-Token'] = self.session_token

        # https://docs.aws.amazon.com/general/latest/gr/sigv4-create-canonical-request.html
        canonical_headers = ''.join('{0}:{1}\n'.format(k.lower(), headers[k]) for k in sorted(headers))
        signed_headers = ';'.join(k.lower() for k in sorted(headers))
        payload_hash = sha256(body.encode('utf-8')).hexdigest()
        canonical_request = '\n'.join([method, '/', '', canonical_headers, signed_headers, payload_hash])

        # https://docs.aws.amazon.com/general/latest/gr/sigv4-create-string-to-sign.html
        algorithm = 'AWS4-HMAC-SHA256'
        credential_scope = '/'.join([timestamp[0:8], 'us-east-1', 'sts', 'aws4_request'])
        canonical_request_hash = sha256(canonical_request.encode('utf-8')).hexdigest()
        string_to_sign = '\n'.join([algorithm, timestamp, credential_scope, canonical_request_hash])

        # https://docs.aws.amazon.com/general/latest/gr/sigv4-calculate-signature.html
        key = 'AWS4{0}'.format(self.secret_key).encode('utf-8')
        key = hmac.new(key, timestamp[0:8].encode('utf-8'), sha256).digest()
        key = hmac.new(key, 'us-east-1'.encode('utf-8'), sha256).digest()
        key = hmac.new(key, 'sts'.encode('utf-8'), sha256).digest()
        key = hmac.new(key, 'aws4_request'.encode('utf-8'), sha256).digest()
        signature = hmac.new(key, string_to_sign.encode('utf-8'), sha256).hexdigest()

        # https://docs.aws.amazon.com/general/latest/gr/sigv4-add-signature-to-request.html
        authorization = '{0} Credential={1}/{2}, SignedHeaders={3}, Signature={4}'.format(
            algorithm, self.access_key, credential_scope, signed_headers, signature)
        headers['Authorization'] = authorization



#!/usr/bin/env python
from setuptools import setup, find_packages

setup(
    name='async-hvac-fork',
    version='0.6.1',
    description='HashiCorp Vault API client, forked from Aloomaio/async-hvac',
    long_description='HashiCorp Vault API python 3.8+ client using asyncio. Supports aiohttp==3.8.1',
    author='northpowered',
    author_email='',
    url='https://github.com/northpowered/async-hvac',
    keywords=['hashicorp', 'vault', 'hvac'],
    classifiers=['License :: OSI Approved :: Apache Software License'],
    packages=find_packages(),
    install_requires=[
        'aiohttp==3.8.1',
    ],
    include_package_data=True,
    package_data={'async_hvac': ['version']},
    extras_require={
        'parser': ['pyhcl==0.3.10']
    }
)
